########################################################################
# Script for generation temperature predicton for each location

# To run the script:
#   python3 main_t2m.py argv1 argv2
#   @param argv1: Classification method -- [RF, MLP. KNN]
#                 Submission: RF
#   @param argv2: "w34" for week 3-4 prediction
#                 "w56" for week 5-6 prediction
######################################################################

import xarray as xr
import numpy as np
import pandas as pd
import sys, pickle, os, re
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import NearestNeighbors
from skimage.measure import block_reduce
from scripts_tp import *

# ---------------------------------------------------------------------

cache_path = "../../data/newdata"
ecmwf_path = "../download_data/data"
v = 'tp'

### argument
CLS = "RF"                                  ### Choose classification method. Suggestion: RF
wk = sys.argv[1]                            ### Choose  week 3-4 or week 5-6
var_name = "v"

# ---------------

flag_model = 1
flag_pred = 1
flag_test2020 = 1

scale = False
n = 10
smallv = 1e-10

time_train_start, time_train_end = '2000', str(int(sys.argv[2])-1)      # train
time_valid_start, time_valid_end = sys.argv[2], sys.argv[3]             # valid
time_test = '2020'                                                      # test

# ---------------------------------------------------------------------
### result saved path
root_dir = (f"../result_sys_valid_tp_ecmwf_tmp1/cv/cv_{time_valid_start}_{time_valid_end}/")
if not os.path.exists(root_dir):
    os.makedirs(root_dir)
print("\n>>>>>>>>>", root_dir, "<<<<<<<<<<<\n")

# ------------------
### run week 3-4 prediction
if wk == "w34":

    flag_w34 = 1                            ### predict w34
    flag_w56 = 0
    
    num_indent = 4

    ### RF hyper parameters
    nest = 100
    ccalpha = 0
    max_depth = 10                          ### 10 for w34, 1 for w56
    criterion = "gini"                      ### entropy for w34, gini for w56

    ### feature parametrs
    flag_past_day = 1
    flag_past_year = 10                     ### whether to use past years' same time value
    one_year_past = 0                       ### if flag_past_year == 1, choose whether only use past one year
    N_year_past = 1-one_year_past           ### if flag_past_year == 1, choose whether only use past N years
    N_years = 1

    flag_ecmwf_anom = 1
    flag_use_w34 = 0

elif wk == "w56":

    flag_w56 = 1                            ### predict w56
    flag_w34 = 0

    num_indent = 6

    ### RF hyper parameters
    nest = 100
    ccalpha = 0
    max_depth = 10                               ### 10 for w34, 1 for w56
    criterion = "entropy"                    ### entropy for w34, gini for w56
    
    ### feature parametrs
    flag_past_day = 1
    flag_past_year = 1                      ### whether to use past years' same time value
    one_year_past = 0                       ### if flag_past_year == 1, choose whether only use past one year
    N_year_past = 1-one_year_past           ### if flag_past_year == 1, choose whether only use past N years
    N_years = 1

    flag_ecmwf_anom = 1
    flag_use_w34 = 1

else:
    print("Wrong week index!!!")

# --------------------------------------------
### train output
try:
    obs_2000_2019 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc")
except:
    obs_2000_2019 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr")
### test output
try:
    obs_2020 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc")
except:
    obs_2020 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr")
### train input
try:
    hind_2000_2019 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc")
except:
    hind_2000_2019 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr")
### test input
try:
    fct_2020 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc")
except:
    fct_2020 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr")

### tercile
tercile_file = f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.nc"
tercile_edges = xr.open_dataset(tercile_file)

# -------------

if flag_w34 == 1:
    obs_2000_2019_w34 = obs_2000_2019.isel(lead_time=[0])
    obs_2020_w34 = obs_2020.isel(lead_time=[0])

    hind_2000_2019_w34 = hind_2000_2019.isel(lead_time=[0])
    fct_2020_w34 = fct_2020.isel(lead_time=[0])
    
    tercile_edges_v = tercile_edges.isel(lead_time=0)[v]

elif flag_w56 == 1:
    obs_2000_2019_w34 = obs_2000_2019.isel(lead_time=[1])
    obs_2020_w34 = obs_2020.isel(lead_time=[1])

    hind_2000_2019_w34 = hind_2000_2019.isel(lead_time=[1])
    fct_2020_w34 = fct_2020.isel(lead_time=[1])

    tercile_edges_v = tercile_edges.isel(lead_time=1)[v]

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = mask.notnull()[v]

### Mask training input
hind_2000_2019_w34_msk = hind_2000_2019_w34.where(obs_2020_w34.isel(forecast_time=0, lead_time=0, drop=True).notnull())

# --------------------------------------------
### averaged removed sequences
train_data = obs_2000_2019_w34.sel(forecast_time=slice(time_train_start, time_train_end))[v].values.transpose(1,0,2,3)
train_data = np.mean(train_data, axis=1)

test_data = obs_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.transpose(1,0,2,3)
test_data = np.mean(test_data, axis=1)

test_2020 = obs_2020_w34.sel(forecast_time=time_test)[v].values.transpose(1,0,2,3)
test_2020 = np.mean(test_2020, axis=1)

if v == 'tp':
    train_data, Orig_1Dplot, After_1Dplot = Fill_Nan_pastInfor(train_data)

train_y = train_data
test_y = test_data
y_2020 = test_2020

train_avg_rm = train_data[num_indent:, :, :]
test_avg_rm = test_data
test_2020_avg_rm = test_2020

train_y_avg_rm = train_avg_rm
test_y_avg_rm = test_avg_rm
test_2020_y_avg_rm = test_2020_avg_rm

if scale:
    print(">>>>>>>> SCALE Sequence")
    scaler = MinMaxScaler()
    train_y_avg_rm_scale = scaler.fit_transform(train_y_avg_rm.reshape(train_y_avg_rm.shape[0], -1)).reshape(train_y_avg_rm.shape)
    test_y_avg_rm_scale = scaler.transform(test_y_avg_rm.reshape(test_y_avg_rm.shape[0], -1)).reshape(test_y_avg_rm.shape)
    test_2020_y_avg_rm_scale = scaler.transform(test_2020_y_avg_rm.reshape(test_2020_y_avg_rm.shape[0], -1)).reshape(test_2020_y_avg_rm.shape)
else:
    print(">>>>>>>> NOT SCALE Sequence")
    train_y_avg_rm_scale = train_y_avg_rm
    test_y_avg_rm_scale = test_y_avg_rm
    test_2020_y_avg_rm_scale = test_2020_y_avg_rm

if v == "t2m":
    mask_flt = np.where(~np.isnan(train_avg_rm[316]))
elif v == "tp":
    mask_flt = np.where(~np.isnan(train_data[0]))
    train_y_avg_rm_scale = np.nan_to_num(train_y_avg_rm_scale)
    test_y_avg_rm_scale = np.nan_to_num(test_y_avg_rm_scale)
    test_2020_y_avg_rm_scale = np.nan_to_num(test_2020_y_avg_rm_scale)

# -------------------------------------------

fct_data = hind_2000_2019_w34.sel(forecast_time=slice(time_train_start, time_train_end))[v].values.squeeze().transpose(1,0,2,3)
if v == "tp":
    fct_data = np.nan_to_num(fct_data)
fct_data_mean = np.mean(fct_data, axis=1)
fct_data_std = np.std(fct_data, axis=1)

fct_test = hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.squeeze().transpose(1,0,2,3)
if v == "tp":
    fct_test = np.nan_to_num(fct_test)
fct_test_mean = np.mean(fct_test, axis=1)
fct_test_std = np.std(fct_test, axis=1)

fct_2020 = fct_2020_w34.sel(forecast_time=time_test)[v].values.squeeze().transpose(1,0,2,3)
if v == "tp":
    fct_2020 = np.nan_to_num(fct_2020)
fct_2020_mean = np.mean(fct_2020, axis=1)
fct_2020_std = np.std(fct_2020, axis=1)

fct_train_mean = fct_data_mean[num_indent:, :, :]
fct_train_std = fct_data_std[num_indent:, :, :]

# ---------------------------------------------------------------
### ecmwf parameters
def cal_anom(parameter, flag_w56):
    dates = xr.cftime_range(start='20000102',freq='7D', periods=53)
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        ecmwf_all = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.nc").isel(lead_time=[int(flag_w56)])
        ecmwf_param = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        ecmwf_2020 = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_forecast_2020_biweekly_agg.nc").isel(lead_time=[int(flag_w56)])
        train_shape = ecmwf_param.to_array().squeeze().shape
        avgs = np.zeros((len(dates),train_shape[-2], train_shape[-1]))
    elif parameter in ["gh", "u", "v", "q"]:
        ecmwf_all = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)])
        ecmwf_param = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        ecmwf_2020 = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_forecast_2020_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)])
        train_shape = ecmwf_param.to_array().squeeze().shape
        avgs = np.zeros((len(dates), train_shape[-3], train_shape[-2], train_shape[-1]))
    
    if parameter in ["sm100", "sm20", "st20"]:
        ecmwf_param = ecmwf_param.mean("depth_below_and_layer")
        ecmwf_test = ecmwf_test.mean("depth_below_and_layer")
        ecmwf_2020 = ecmwf_2020.mean("depth_below_and_layer")
    if parameter in ["tcc", "tcw"]:
        ecmwf_param = ecmwf_param.mean("entireAtmosphere")
        ecmwf_test = ecmwf_test.mean("entireAtmosphere")
        ecmwf_2020 = ecmwf_2020.mean("entireAtmosphere")
    if parameter == "msl":
        ecmwf_param = ecmwf_param.mean("meanSea")
        ecmwf_test = ecmwf_test.mean("meanSea")
        ecmwf_2020 = ecmwf_2020.mean("meanSea")
    for i,f in enumerate(dates):
        date_range = pd.date_range(start=f.strftime('%Y-%m-%d'),
                    end=f'{time_train_end}-' +  f'{f.month:02d}' + '-' + str(f.day), freq=pd.DateOffset(years=1)).to_list()
        avgs[i] = ecmwf_param.sel(forecast_time=date_range).mean('realization').mean('forecast_time').to_array().squeeze()
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        avgs_ext = np.tile(avgs.transpose([1,2,0]), int(int(time_train_end)-int(time_train_start)+1)).transpose([2,0,1])
        avgs_val = np.tile(avgs.transpose([1,2,0]), int(int(time_valid_end)-int(time_valid_start)+1)).transpose([2,0,1])
    elif parameter in ["gh", "u", "v", "q"]:
        avgs_ext = np.tile(avgs.transpose([1,2,3,0]), int(int(time_train_end)-int(time_train_start)+1)).transpose([3,0,1,2])
        avgs_val = np.tile(avgs.transpose([1,2,3,0]), int(int(time_valid_end)-int(time_valid_start)+1)).transpose([3,0,1,2])
    anom_train = ecmwf_param.mean('realization') - avgs_ext
    anom_test = ecmwf_test.mean('realization') - avgs_val
    anom_2020 = ecmwf_2020.mean('realization') - avgs
    anom_train, anom_test = anom_train.to_array().squeeze(), anom_test.to_array().squeeze()
    anom_2020 = anom_2020.to_array().squeeze()
    anom_train = anom_train[num_indent:]
    return anom_train.values, anom_test.values, anom_2020.values

def load_ecmwf(parameter, flag_w56):
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        ecmwf_all = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.nc").isel(lead_time=[int(flag_w56)]).mean('realization')
        ecmwf_train = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        ecmwf_2020 = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_forecast_2020_biweekly_agg.nc").isel(lead_time=[int(flag_w56)]).mean('realization')
    elif parameter in ["gh", "u", "v", "q"]:
        ecmwf_all = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)]).mean('realization')
        ecmwf_train = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        ecmwf_2020 = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_forecast_2020_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)]).mean('realization')

    if parameter in ["sm100", "sm20", "st20"]:
        ecmwf_train = ecmwf_train.mean("depth_below_and_layer")
        ecmwf_test = ecmwf_test.mean("depth_below_and_layer")
        ecmwf_2020 = ecmwf_2020.mean("depth_below_and_layer")
    if parameter == "msl":
        ecmwf_param = ecmwf_param.mean("meanSea")
        ecmwf_test = ecmwf_test.mean("meanSea")
        ecmwf_2020 = ecmwf_2020.mean("meanSea")
    if parameter in ["tcc", "tcw"]:
        ecmwf_param = ecmwf_param.mean("entireAtmosphere")
        ecmwf_test = ecmwf_test.mean("entireAtmosphere")
        ecmwf_2020 = ecmwf_2020.mean("entireAtmosphere")
    ecmwf_train, ecmwf_test = ecmwf_train.to_array().squeeze(), ecmwf_test.to_array().squeeze()
    ecmwf_2020 = ecmwf_2020.to_array().squeeze()
    train_set, test_set = ecmwf_train[num_indent:], ecmwf_test
    return train_set.values, test_set.values, ecmwf_2020.value

# ---------------------------------------------------------------

def form_ds(pred_test, time, ds_cmpr, tercile_edges, mask):

    pred_test = np.expand_dims(pred_test, axis=1)

    ds_cmpr = ds_cmpr.sel(forecast_time=time)
    pred_test_da = xr.DataArray(pred_test,
                            dims=['forecast_time', 'lead_time', 'latitude', 'longitude'],
                            coords={'forecast_time': ds_cmpr.forecast_time,
                                    'lead_time': ds_cmpr.lead_time,
                                    'latitude': ds_cmpr.latitude,
                                    'longitude': ds_cmpr.longitude},
                                    )

    preds_test_add = add_valid_time_from_forecast_reference_time_and_lead_time(pred_test_da)
    preds_test_add = preds_test_add.to_dataset(name=v)
    
    preds_test_p = make_probabilistic(preds_test_add.expand_dims('realization'), tercile_edges, mask=mask)

    # -----
    ds = add_year_week_coords(preds_test_add.expand_dims('realization'))
    groupby_coord = 'week'
    tercile_edges_v_1 = tercile_edges.sel({groupby_coord: ds.coords[groupby_coord]}).values

    return preds_test_p, tercile_edges_v_1

### -----
def read_monthly(name, data_flag='nino'):

    if data_flag == 'nino':
        path = '../data/nino/'
        data =  pd.read_csv(path+name, sep="\n")
        data = data[:151]
        foo = lambda x: pd.Series([i for i in x.split('   ')])

    if data_flag == 'oscil':
        path = '../data/Oscillation/'
        data =  pd.read_csv(path+name, sep="\n", header=None)
        data = data[:-1]
        foo = lambda x: pd.Series([i for i in x.split('  ')])

    data = data.iloc[:,0].apply(foo)
    data.columns = [str(i) for i in range(13)]
    data = data.set_index('0')
    data_cut = data.iloc[-21:, :]

    df = pd.DataFrame()
    n_4 = [2,3,5,6,8,9,11]
    n_5 = [1,4,7,10,12]

    for col in data_cut.columns:
        if int(col) in n_4:
            p4 = pd.concat([data_cut[col]] * (4), axis=1, ignore_index=True)
            p4.columns = [col+"_1", col+"_2", col+"_3", col+"_4"]
            df = pd.concat([df, p4], axis=1)
        elif int(col) in n_5:
            p5 = pd.concat([data_cut[col]] * (5), axis=1, ignore_index=True)
            p5.columns = [col+"_1", col+"_2", col+"_3", col+"_4", col+"_5"]
            df = pd.concat([df, p5], axis=1)
    
    df_data = df[:21]
    df_data_list = []

    for i in range(df_data.shape[0]):
        df_data_list.extend(df_data.iloc[i, :])
    df_data_list = np.array(df_data_list)

    df_data_list = df_data_list[:-num_indent]

    df_train = df_data_list[:len(train_avg_rm)]
    df_test = df_data_list[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]
    df_2020 = df_data_list[-len(test_2020_avg_rm):]

    df_train = df_train.astype(float)
    df_test = df_test.astype(float)
    df_2020 = df_2020.astype(float)

    return df_train, df_test, df_2020


### -----

def read_daily(df):

    df_lst = []
    dates = xr.cftime_range(start='20000102',freq='7D', periods=53)
    for row_idx in range(df.shape[0]):
        row = df.iloc[row_idx, :]
        mon, day = int(row["MM"]), int(row["DD"])
        
        for i, f in enumerate(dates):
            if (mon == f.month) and (day == f.day):
                tmp = df.iloc[row_idx:row_idx+14, -1].mean()
                df_lst.append(tmp)
            
    df_lst = np.array(df_lst)
    df_lst = df_lst[:-num_indent]

    df_train = df_lst[:len(train_avg_rm)]
    df_test = df_lst[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]
    df_2020 = df_lst[-len(test_2020_avg_rm):]

    return df_train, df_test, df_2020

### -----

def read_rmm(df):
    dates = xr.cftime_range(start='20000102',freq='7D', periods=53)
    df_lst = []
    for row_idx in range(df.shape[0]):
        row = df.iloc[row_idx, :]
        mon, day = int(row["MM"]), int(row["DD"])
        
        for i, f in enumerate(dates):
            if (mon == f.month) and (day == f.day):
                tmp = df.iloc[row_idx-N_rmm_pst:row_idx, -1].values
                df_lst.append(tmp)
    if flag_w34 == 1:
        df_lst = df_lst[2:-2]
    else:
        df_lst = df_lst[2:-4]
    df_lst = np.array(df_lst)

    df_train = df_lst[:len(train_avg_rm)]
    df_test = df_lst[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]
    df_2020 = df_lst[-len(test_2020_avg_rm):]

    return df_train, df_test, df_2020

### -----

def biw_varb(path):
    root_dir = "../data/biw_stat_"+v+"_"+wk+"/"
    data = np.load(root_dir+path)
    data_all = data

    for t in range(20):
        data_all = np.concatenate([data_all, data], axis=0)
    data_all = data_all[num_indent:, :, :]

    data_train = data_all[:len(train_avg_rm), :, :]
    data_test = data_all[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm), :, :]
    data_2020 = data_all[-len(test_2020_avg_rm):, :, :]

    return data_train, data_test, data_2020

#################################################################
print(">>>>>> Loading Features <<<<<<<<")
train_seq_p, train_terc = form_ds(train_y, slice(time_train_start,time_train_end), 
                                    hind_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end)),
                                    tercile_edges_v, mask_v)
test_seq_p, test_terc = form_ds(test_y, slice(time_valid_start, time_valid_end), 
                                hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end)),
                                tercile_edges_v, mask_v)
test_2020_seq_p, test_2020_terc = form_ds(y_2020, time_test, fct_2020_w34,
                                            tercile_edges_v, mask_v)                            
# -------------------------------------------

obs_2000_2019_w34_add = add_year_week_coords(obs_2000_2019_w34)
hind_2000_2019_w34_add = add_year_week_coords(hind_2000_2019_w34)
fct_2020_w34_add = add_year_week_coords(fct_2020_w34)

hind_train_add = hind_2000_2019_w34_add.sel(forecast_time=slice(time_train_start,time_train_end))
hind_test_add = hind_2000_2019_w34_add.sel(forecast_time=slice(time_valid_start, time_valid_end))
obs_train_add = obs_2000_2019_w34_add.sel(forecast_time=slice(time_train_start,time_train_end))
obs_test_add = obs_2000_2019_w34_add.sel(forecast_time=slice(time_valid_start, time_valid_end))

bias_mean_2000_2018 = (hind_train_add.mean('realization') - obs_train_add).groupby('week').mean().compute()
bias_std_2000_2018 = (hind_train_add.mean('realization') - obs_train_add).groupby('week').std().compute()

bias_mean_train = bias_mean_2000_2018.sel(week=hind_train_add.week)[v].squeeze().values
bias_std_train = bias_std_2000_2018.sel(week=hind_train_add.week)[v].squeeze().values

bias_mean_test = bias_mean_2000_2018.sel(week=hind_test_add.week)[v].squeeze().values
bias_std_test = bias_std_2000_2018.sel(week=hind_test_add.week)[v].squeeze().values

bias_mean_2020 = bias_mean_2000_2018.sel(week=fct_2020_w34_add.week)[v].squeeze().values
bias_std_2020 = bias_std_2000_2018.sel(week=fct_2020_w34_add.week)[v].squeeze().values

bias_mean_train = bias_mean_train[num_indent:, :, :]
bias_std_train = bias_std_train[num_indent:, :, :]

# -------
### El NINO
nin12_train, nin12_test, nin12_2020 = read_monthly("nino12.long.anom.data")
nin34_train, nin34_test, nin34_2020 = read_monthly("nino34.long.anom.data")

# -------
### Oscillation
nao_train, nao_test, nao_2020 = read_monthly("norm.nao.monthly.b5001.current.ascii.table.txt", data_flag='oscil')
pdo_train, pdo_test, pdo_2020 = read_monthly("ersst.v5.pdo.dat.txt", data_flag='oscil')

# -------
### biweekly statistics
avg_train, avg_test, avg_2020 = biw_varb("biweekly_historical_avgs.npy")
std_train, std_test, std_2020 = biw_varb("biweekly_historical_std.npy")
med_train, med_test, med_2020 = biw_varb("biweekly_historical_med.npy")
skw_train, skw_test, skw_2020 = biw_varb("biweekly_historical_skew.npy")
krt_train, krt_test, krt_2020 = biw_varb("biweekly_historical_kurt.npy")

# -------
### Sunspot Cycles

n_4 = [2,3,5,6,8,9,11]
n_5 = [1,4,7,10,12]

solar = pd.read_json("../data/observed-solar-cycle-indices.json")
solar = solar[solar['time-tag'] >= '2000-01']
solar = solar[solar['time-tag'] <= '2020-12']
solar_ssn = solar[["time-tag", "ssn", "smoothed_ssn"]]

# -------
### Madden-Julian

path = '../data/Oscillation/rmm.txt'
data =  pd.read_csv(path, sep="\n")
foo = lambda x: pd.Series([i for i in x.split('\t')])
data = data.iloc[:,0].apply(foo)
data.columns = ["YYYY", "MM", "DD", "RMM1", "RMM2", "phase", "amplitude", "Missing Value= 1.E36 or 999"]
data = data.drop(["Missing Value= 1.E36 or 999"], axis=1)
data_md = data[data["YYYY"] >= '2000']
data_md = data_md[data_md["YYYY"] <= '2020']

data_md = data_md.astype(float)
data_rmm1 = data_md[['YYYY', 'MM', 'DD', 'RMM1']]
data_rmm2 = data_md[['YYYY', 'MM', 'DD', 'RMM2']]
data_amp = data_md[['YYYY', 'MM', 'DD', 'amplitude']]


# -------
### Great Lakes Ice Cover
path = '../data/great_lake/N_seaice_extent_daily_v3.0.csv'
data_gl =  pd.read_csv(path, header=1)
data_gl = data_gl.iloc[: , :-2]
data_gl.columns = ['YYYY', 'MM', 'DD', '10^6 sq km']
data_gl = data_gl[data_gl["YYYY"]>=2000]
data_gl = data_gl[data_gl["YYYY"]<=2020]
gl_train, gl_test, gl_2020 = read_daily(data_gl)

# -------
### QBO
n_4 = [2,3,5,6,8,9,11]
n_5 = [1,4,7,10,12]

path = "../data/Oscillation/qbo.dat"
pdo = pd.read_csv(path, sep="\n", header=None)
pdo = pdo[7+(12*(2000-1953)):-11]
foo = lambda x: pd.Series([t for t in [i for i in x.split(' ')] if t])
pdo = pdo.iloc[:,0].apply(foo).iloc[:,1:]
pdo.columns = ["YYMM", "70hPaN", "50hPaN", "40hPaN", "30hPaN", "20hPaN", "15hPaN", "10hPaN"]
pdo[['YY','MM']] = pdo.YYMM.apply(lambda x: pd.Series([x[index : index+2] for index in range(0, len(x), 2)]))
pdo = pdo.iloc[:,1:]

a = [i+1 for i in range(12)]
rep = [4 if i in n_4 else 5 for i in a]*21
pdo["times"] = rep

pdo_repeat = pdo.loc[pdo.index.repeat(pdo.times)]
pdo_repeat = pdo_repeat.iloc[:-num_indent, :]

qbo_train = pdo_repeat.iloc[:len(train_avg_rm), :-3]
qbo_test = pdo_repeat.iloc[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm), :-3]
qbo_2020 = pdo_repeat.iloc[-len(test_2020_avg_rm):, :-3]
qbo_train, qbo_test = qbo_train.astype(float), qbo_test.astype(float)
qbo_2020 = qbo_2020.astype(float)

# -------
### Another variable (tp or t2m)
if v == "tp": 
    v_other = "t2m"
elif v == "t2m":
    v_other = "tp"

# flag_another
train_other_tmp = obs_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end))[v_other].values.transpose(1,0,2,3)
train_other_tmp = np.mean(train_other_tmp, axis=1)
train_other = train_other_tmp[:-num_indent, :, :]

test_other_tmp = obs_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v_other].values.transpose(1,0,2,3)
test_other_tmp = np.mean(test_other_tmp, axis=1)
test_other = np.concatenate([train_other[-num_indent:, :, :], test_other_tmp[:-num_indent, :, :]])

test_2020_other_tmp = obs_2020_w34.sel(forecast_time=time_test)[v_other].values.transpose(1,0,2,3)
test_2020_other_tmp = np.mean(test_2020_other_tmp, axis=1)
test_2020_other = np.concatenate([train_other[-num_indent:, :, :], test_2020_other_tmp[:-num_indent, :, :]])

# flag_another_fct
train_other_fct = hind_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end))[v_other].values.squeeze().transpose(1,0,2,3)
train_other_fct = np.nan_to_num(train_other_fct)
train_other_fct = np.mean(train_other_fct, axis=1)
train_other_fct = train_other_fct[num_indent:, :, :]

test_other_fct = hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v_other].values.squeeze().transpose(1,0,2,3)
test_other_fct = np.nan_to_num(test_other_fct)
test_other_fct = np.mean(test_other_fct, axis=1)

test_2020_other_fct = fct_2020_w34.sel(forecast_time=time_test)[v_other].values.squeeze().transpose(1,0,2,3)
test_2020_other_fct = np.nan_to_num(test_2020_other_fct)
test_2020_other_fct = np.mean(test_2020_other_fct, axis=1)

# # ---------------------------------------------------------------
# w34_pred
if (flag_w56 == 1) and (flag_use_w34 == 1):
    w34_train = np.load("../result_tp_ecmwf/RegModel_RF_gini_w341_w560_ecmwf11_rsn1sm1000msl1st200tcc1tcw1sm201ci0_sst00-1gh18u10v16q1-1/train_prob_AR_10.npz")["arr_0"]
    w34_test = np.load("../result_tp_ecmwf/RegModel_RF_gini_w341_w560_ecmwf11_rsn1sm1000msl1st200tcc1tcw1sm201ci0_sst00-1gh18u10v16q1-1/test_prob_AR_10.npz")["arr_0"]
    w34_2020 = np.load("../result_tp_ecmwf/RegModel_RF_gini_w341_w560_ecmwf11_rsn1sm1000msl1st200tcc1tcw1sm201ci0_sst00-1gh18u10v16q1-1/test_2020_prob_AR_10.npz")["arr_0"]
    w34_pred_train = np.nan_to_num(w34_train.transpose(1,0,2,3))
    cc = np.zeros([(avg_train.shape[0]-w34_pred_train.shape[0]), w34_pred_train.shape[1], w34_pred_train.shape[2], w34_pred_train.shape[3]])
    w34_pred_train = np.concatenate([w34_pred_train, cc])
    w34_pred_test = np.nan_to_num(w34_test.transpose(1,0,2,3))
    w34_pred_2020 = np.nan_to_num(w34_2020.transpose(1,0,2,3))

    # w34_fct
    w34_fct_train = hind_2000_2019.isel(lead_time=[0]).sel(forecast_time=slice(time_train_start, time_train_end))[v].values.squeeze().transpose(1,0,2,3)
    w34_fct_test = hind_2000_2019.isel(lead_time=[0]).sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.squeeze().transpose(1,0,2,3)
    w34_fct_2020 = fct_2020.isel(lead_time=[0]).sel(forecast_time=time_test)[v].values.squeeze().transpose(1,0,2,3)
    w34_fct_train = np.mean(np.nan_to_num(w34_fct_train), axis=1)[num_indent:, :, :]
    w34_fct_test = np.mean(np.nan_to_num(w34_fct_test), axis=1)
    w34_fct_2020 = np.mean(np.nan_to_num(w34_fct_2020), axis=1)

# ---------------------------------------------------------------
if flag_ecmwf_anom == 0 :
    ecmwf_rsn_train, ecmwf_rsn_test, ecmwf_rsn_2020 = load_ecmwf("rsn", flag_w56)
    ecmwf_sm100_train, ecmwf_sm100_test, ecmwf_sm100_2020 = load_ecmwf("sm100", flag_w56)
    ecmwf_sst_train, ecmwf_sst_test, ecmwf_sst_2020 = load_ecmwf("sst", flag_w56)
    ecmwf_gh_train, ecmwf_gh_test, ecmwf_gh_2020 = load_ecmwf("gh", flag_w56)
    ecmwf_ci_train, ecmwf_ci_test, ecmwf_ci_2020 = load_ecmwf("ci", flag_w56)
    ecmwf_u_train, ecmwf_u_test, ecmwf_u_2020 = load_ecmwf("u", flag_w56)
    ecmwf_v_train, ecmwf_v_test, ecmwf_v_2020 = load_ecmwf("v", flag_w56)
    ecmwf_msl_train, ecmwf_msl_test, ecmwf_msl_2020 = load_ecmwf("msl", flag_w56)
    ecmwf_st20_train, ecmwf_st20_test, ecmwf_st20_2020 = load_ecmwf("st20", flag_w56)
    ecmwf_tcc_train, ecmwf_tcc_test, ecmwf_tcc_2020 = load_ecmwf("tcc", flag_w56)
    ecmwf_tcw_train, ecmwf_tcw_test, ecmwf_tcw_2020 = load_ecmwf("tcw", flag_w56)
    ecmwf_sm20_train, ecmwf_sm20_test, ecmwf_sm20_2020 = load_ecmwf("sm20", flag_w56)
    ecmwf_q_train, ecmwf_q_test, ecmwf_q_2020 = load_ecmwf("q", flag_w56)
    
elif flag_ecmwf_anom == 1:
    ecmwf_rsn_train, ecmwf_rsn_test, ecmwf_rsn_2020 = cal_anom("rsn", flag_w56)
    ecmwf_sm100_train, ecmwf_sm100_test, ecmwf_sm100_2020 = cal_anom("sm100", flag_w56)
    ecmwf_sst_train, ecmwf_sst_test, ecmwf_sst_2020 = cal_anom("sst", flag_w56)
    ecmwf_gh_train, ecmwf_gh_test, ecmwf_gh_2020 = cal_anom("gh", flag_w56)
    ecmwf_ci_train, ecmwf_ci_test, ecmwf_ci_2020 = cal_anom("ci", flag_w56)
    ecmwf_u_train, ecmwf_u_test, ecmwf_u_2020 = cal_anom("u", flag_w56)
    ecmwf_v_train, ecmwf_v_test, ecmwf_v_2020 = cal_anom("v", flag_w56)
    ecmwf_msl_train, ecmwf_msl_test, ecmwf_msl_2020 = cal_anom("msl", flag_w56)
    ecmwf_st20_train, ecmwf_st20_test, ecmwf_st20_2020 = cal_anom("st20", flag_w56)
    ecmwf_tcc_train, ecmwf_tcc_test, ecmwf_tcc_2020 = cal_anom("tcc", flag_w56)
    ecmwf_tcw_train, ecmwf_tcw_test, ecmwf_tcw_2020 = cal_anom("tcw", flag_w56)
    ecmwf_sm20_train, ecmwf_sm20_test, ecmwf_sm20_2020 = cal_anom("sm20", flag_w56)
    ecmwf_q_train, ecmwf_q_test, ecmwf_q_2020 = cal_anom("q", flag_w56)        

print(">>>>>> Finish Loading Features <<<<<<<<")
##################################################################

df_path = f"../result_sys_valid_tp_ecmwf_tmp1/level_var/{var_name}_score_set.csv"
df_level = pd.read_csv(df_path, header=None, index_col=0)

df_all_max = pd.read_csv("../result_sys_valid_tp_ecmwf_tmp1/score_set.csv", header=None)
df_all_max = df_all_max.set_index(0)

region = pd.read_csv('../data/Regions.txt', sep=',')
d_region = {}
for i in range(region.shape[0]):
    d_region[region.iloc[i]["Region"]] = d_region.get(region.iloc[i]["Region"], [])+[list(region.iloc[i, 1:])]

# -------------
if flag_model == 1:
    for row in range(df_all_max.shape[0]):
        name = df_all_max.iloc[[row]].index.values[0]   
        print("!!!!", name)

        best_set_str = df_all_max.loc[[name]][2].values[0]
        unit_fldr = df_all_max.loc[[name]][1].values[0]

        ### flag
        best_set = best_set_str.split("_")
        flag_fct_bias = int(best_set[0][-1])

        flag_biw = [int(best_set[1][3]), int(best_set[1][4]), int(best_set[1][5]), int(best_set[1][6]), int(best_set[1][7])]                   ### whether to use biweekly statistics, [avg, std, skew, kurt, median]

        flag_nino = int(best_set[2][-1])
        flag_solar = int(best_set[3][-1])                         ### Sunspot Cycles [0: not use; 1: normal; 2: smoothed]
        flag_nao = int(best_set[4][-1])                           ### NAO data (North Atlantic Oscillation) [0: not use; 1: use]
        flag_pdo = int(best_set[5][-1])                          ### PDO data (Pacific Decadal Oscillation) [0: not use; 1: use]
        flag_gl = int(best_set[6][-1])                           ### Great Lakes Ice Cover [0: not use; 1: use]

        flag_qbo = 0
        flag_qbo_col = -2

        flag_rmm = 0                            ### whether to use Madden-Julian Oscillation RMM1 & RMM2 value
        flag_rmm_avg = 0
        N_rmm_pst = 14
        flag_amp = 0                           ### whether to use Madden-Julian Oscillation amplitude value

        flag_another = int(best_set[8][-2])                       ### another variable (tp or t2m)
        flag_another_fct = int(best_set[8][-1])

        flag_use_w34 = 0
        w34_pred = 0
        w34_fct = 0

        # ---------
        flag_use_ecmwf = 1
        flag_ecmwf_anom = 1

        ecmwf_rsn = int(best_set[10][-1])
        ecmwf_sm100 = int(best_set[11][-1])
        ecmwf_sm20 = int(best_set[16][-1])

        ecmwf_msl = int(best_set[12][-1])
        ecmwf_st20 = int(best_set[13][-1])
        ecmwf_tcc = int(best_set[14][-1])
        ecmwf_tcw = int(best_set[15][-1])
        ecmwf_ci = int(best_set[17][-1])

        ecmwf_sst = 0
        n_neig = 0
        sst_reduce = -1

        ecmwf_gh = 0
        gh_col = -2

        ecmwf_u = 0
        u_col = -2
        ecmwf_v = 0
        v_col = -2

        ecmwf_q = 0
        q_col = -2

        ### level features
        if df_level.loc[[name]][2].values[0].startswith("qbo"):
            level_lst = df_level.loc[[name]][2].values[0].split("_")
            level_unit = df_level.loc[[name]][1].values[0].split("_")[-1]

            flag_qbo = int(level_lst[0][3])
            flag_qbo_col = int(level_lst[0][4:])
            ecmwf_gh = int(level_lst[1][2])
            gh_col = int(level_lst[1][3:])
            ecmwf_u = int(level_lst[2][1])
            u_col = int(level_lst[2][2:])
            ecmwf_v = int(level_lst[3][1])
            v_col = int(level_lst[3][2:])
            ecmwf_q = int(level_lst[4][1])
            q_col = int(level_lst[4][2:])
        
        ### save setting
        d_set = {
                 "flag_fct_bias" : flag_fct_bias, 
                 "flag_biw" : flag_biw, 
                 "flag_nino" : flag_nino, 
                 "flag_solar" : flag_solar, 
                 "flag_nao" : flag_nao, 
                 "flag_pdo" : flag_pdo, 
                 "flag_gl" : flag_gl, 
                 "flag_qbo" : flag_qbo, 
                 "flag_qbo_col" : flag_qbo_col, 
                 "flag_rmm" : flag_rmm, 
                 "flag_rmm_avg" : flag_rmm_avg, 
                 "N_rmm_pst" : N_rmm_pst, 
                 "flag_amp" : flag_amp, 
                 "flag_another" : flag_another, 
                 "flag_another_fct" : flag_another_fct, 
                 "flag_use_w34" : flag_use_w34, 
                 "w34_pred" : w34_pred, 
                 "w34_fct" : w34_fct, 
                 "flag_use_ecmwf" : flag_use_ecmwf, 
                 "flag_ecmwf_anom" : flag_ecmwf_anom, 
                 "ecmwf_rsn" : ecmwf_rsn, 
                 "ecmwf_sm100" : ecmwf_sm100, 
                 "ecmwf_sm20" : ecmwf_sm20, 
                 "ecmwf_msl" : ecmwf_msl, 
                 "ecmwf_st20" : ecmwf_st20, 
                 "ecmwf_tcc" : ecmwf_tcc, 
                 "ecmwf_tcw" : ecmwf_tcw, 
                 "ecmwf_ci" : ecmwf_ci, 
                 "ecmwf_sst" : ecmwf_sst, 
                 "n_neig" : n_neig, 
                 "sst_reduce" : sst_reduce, 
                 "ecmwf_gh" : ecmwf_gh, 
                 "gh_col" : gh_col, 
                 "ecmwf_u" : ecmwf_u, 
                 "u_col" : u_col, 
                 "ecmwf_v" : ecmwf_v, 
                 "v_col" : v_col, 
                 "ecmwf_q" : ecmwf_q, 
                 "q_col" : q_col
                }
        save_set_path = root_dir+'/setting/'
        if not os.path.exists(save_set_path):
            os.makedirs(save_set_path)
        d_name = save_set_path+'/setting_'+name+'.pkl'
        pickle.dump(d_set, open(d_name, 'wb'))

        # ---------------------------------------------------------------
        ## processing to some features
        if ecmwf_sst == 1:
            if n_neig != 0:
                mask_sst = np.where(~np.isnan(ecmwf_sst_train[0]))
                mask_sst_flt = [[mask_sst[0][i],mask_sst[1][i]] for i in range(len(mask_sst[0]))]
                neigh = NearestNeighbors(n_neighbors=2, radius=120)
                neigh.fit(mask_sst_flt)
            if sst_reduce == 1:
                img_train = ecmwf_sst_train
                img_test = ecmwf_sst_test
                img_2020 = ecmwf_sst_2020
                sst_train = block_reduce(img_train, (1,7,7), np.mean, cval='nan')
                sst_test = block_reduce(img_test, (1,7,7), np.mean, cval='nan')
                sst_2020 = block_reduce(img_2020, (1,7,7), np.mean, cval='nan')
            elif sst_reduce == 0:
                sst_train = ecmwf_sst_train
                sst_test = ecmwf_sst_test
                sst_2020 = ecmwf_sst_2020

        if flag_solar != 0: 
            solar_list = []
            for rol in range(len(solar_ssn)):
                ssn = solar_ssn.iloc[rol, flag_solar]
                mm = int(solar_ssn.iloc[rol, 0].split('-')[-1])

                if mm in n_4:
                    solar_list.extend([ssn]*4)
                elif mm in n_5:
                    solar_list.extend([ssn]*5)

            solar_list = np.array(solar_list)
            solar_list = solar_list[:-num_indent]

            solar_train = solar_list[:len(train_avg_rm)]
            solar_test = solar_list[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]
            solar_2020 = solar_list[-len(test_2020_avg_rm):]

        if (flag_rmm == 1) or (flag_rmm_avg == 1):
            rmm1_train, rmm1_test, rmm1_2020 = read_daily(data_rmm1)
            rmm2_train, rmm2_test, rmm2_2020 = read_daily(data_rmm2)
            amp_train, amp_test, amp_2020 = read_daily(data_amp)
        elif (flag_rmm == 1) or (flag_rmm_avg == 0):
            rmm1_train, rmm1_test, rmm1_2020 = read_rmm(data_rmm1)
            rmm2_train, rmm2_test, rmm2_2020 = read_rmm(data_rmm2)

        # ---------------------------------------------------------------
        ### get coord for each region
        loc_region = d_region[name]
        u1, d1, l1, r1 = loc_region[0]
        u2, d2, l2, r2 = u1, d1, l1, r1
        if len(loc_region) > 1:
            u2, d2, l2, r2 = loc_region[1]

        ### loop over loc
        X_train_all = []
        y_train_all = []

        for idx in range(len(mask_flt[0])):

            if idx % 1000 == 0:
                print("Training", idx)

            i = mask_flt[0][idx]
            j = mask_flt[1][idx]  

            ### check whether loc in region
            if ((i>=u1) and (i<d1) and (j>=l1) and (j<r1)) or ((i>=u2) and (i<d2) and (j>=l2) and (j<r2)):

                window = list(np.arange(num_indent, num_indent+9))
                lower = train_terc[num_indent:,0,i,j]
                upper = train_terc[num_indent:,1,i,j]

                ### minmax train sequence
                train_y_avg_rm_scale_mm = ((train_y_avg_rm_scale[:,i,j]-lower)/((upper-lower)+smallv))
                fct_train_mean_part = (fct_train_mean[:,i, j]-lower)/((upper-lower)+smallv)
                fct_train_std_part = (fct_train_std[:,i, j]-lower)/((upper-lower)+smallv)

                ### lag value
                pst_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], len(window)])
                for lgth_idx in range(len(window)):
                    rll = np.roll(train_y_avg_rm_scale_mm, window[lgth_idx])
                    pst_part[:, lgth_idx] = rll

                ### past N year value
                if flag_past_year == 1:
                    ## one year
                    if one_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], 1])
                        rll = np.roll(train_y_avg_rm_scale_mm, 53)
                        year_part[:, 0] = rll

                    ## N years
                    if N_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], N_years])
                        for N_year in range(N_years):
                            rll_N = np.roll(train_y_avg_rm_scale_mm, 53*(N_year+1))
                            year_part[:, N_year] = rll_N
                
                #-----------------------------------------------------------------------------------------------    
                ### Training data
                if flag_past_day == 1:
                    exog_set_train = np.concatenate([pst_part, 
                                                        fct_train_mean_part.reshape(-1,1), 
                                                        fct_train_std_part.reshape(-1,1)], axis=1)
                else:
                    exog_set_train = np.concatenate([fct_train_mean_part.reshape(-1,1), 
                                                        fct_train_std_part.reshape(-1,1)], axis=1)
                if flag_past_year == 1:
                    exog_set_train = np.concatenate([exog_set_train, year_part], axis=1)
                
                if flag_fct_bias == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        bias_mean_train[:, i, j].reshape(-1,1), 
                                                        bias_std_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_use_w34 == 1:
                    if w34_pred == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        w34_pred_train[:,:,i,j]], axis=1)                
                    if w34_fct == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            w34_fct_train[:,i,j].reshape(-1,1)], axis=1) 
                #-----------------------------------------------------------------------------------------------    
                if flag_biw[0] == 1:
                    exog_set_train = np.concatenate([exog_set_train, avg_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[1] == 1:
                    exog_set_train = np.concatenate([exog_set_train, std_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[2] == 1:
                    exog_set_train = np.concatenate([exog_set_train, skw_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[3] == 1:
                    exog_set_train = np.concatenate([exog_set_train, krt_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[4] == 1:
                    exog_set_train = np.concatenate([exog_set_train, med_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_nino == 1:                    
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        nin12_train.reshape(-1,1), nin34_train.reshape(-1,1)], axis=1)
                if flag_solar == 1:
                    exog_set_train = np.concatenate([exog_set_train, solar_train.reshape(-1,1)], axis=1)
                if flag_nao == 1:
                    exog_set_train = np.concatenate([exog_set_train, nao_train.reshape(-1,1)], axis=1)
                if flag_pdo == 1:
                    exog_set_train = np.concatenate([exog_set_train, pdo_train.reshape(-1,1)], axis=1)
                if flag_gl == 1:
                    exog_set_train = np.concatenate([exog_set_train, gl_train.reshape(-1,1)], axis=1)

                if flag_qbo == 1:
                    if flag_qbo_col == -1:
                        exog_set_train = np.concatenate([exog_set_train, qbo_train], axis=1)
                    else:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            qbo_train.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_rmm == 1:
                    if flag_rmm_avg == 1:
                        exog_set_train = np.concatenate([exog_set_train, rmm1_train.reshape(-1,1)], axis=1)
                        exog_set_train = np.concatenate([exog_set_train, rmm2_train.reshape(-1,1)], axis=1)
                    else:
                        exog_set_train = np.concatenate([exog_set_train, rmm1_train], axis=1)
                        exog_set_train = np.concatenate([exog_set_train, rmm2_train], axis=1)
                if flag_amp == 1:
                    exog_set_train = np.concatenate([exog_set_train, amp_train.reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_another == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        train_other[:, i, j].reshape(-1,1)], axis=1)
                if flag_another_fct == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        train_other_fct[:, i, j].reshape(-1,1)], axis=1)    
                #-----------------------------------------------------------------------------------------------    
                if flag_use_ecmwf == 1:
                    if ecmwf_rsn == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_rsn_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm100 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_sm100_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sst == 1:
                        ### k-nearest-neighbour method
                        if n_neig != 0:
                            tt = neigh.kneighbors([[i,j]], n_neig, return_distance=False)
                            sst_seq = ecmwf_sst_train[:, mask_sst[0][tt][0], mask_sst[1][tt][0]].squeeze()
                            if n_neig == 1:
                                sst_seq = sst_seq.reshape(-1,1)
                        ### correlation method
                        if sst_reduce in [0, 1]:
                            a = (sst_train-np.mean(sst_train, axis=0)).transpose([1,2,0])
                            b = (train_y_avg_rm_scale[:,i,j]-train_y_avg_rm_scale[:,i,j].mean()).reshape(-1,1)
                            sst_cor = np.dot(a, b).squeeze() / (np.std(sst_train, axis=0)*np.std(train_y_avg_rm_scale[:,i,j])*sst_train.shape[0])
                            sst_cor_loc = np.unravel_index(np.nanargmax(sst_cor), sst_cor.shape)
                            sst_seq = sst_train[:,sst_cor_loc[0], sst_cor_loc[1]].reshape(-1,1)
                        exog_set_train = np.concatenate([exog_set_train, sst_seq], axis=1)
                    if ecmwf_gh == 1:
                        if gh_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_gh_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_gh_train[:,gh_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_ci == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_ci_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_u == 1:
                        if u_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_u_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_u_train[:,u_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_v == 1:
                        if v_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_v_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_v_train[:,v_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_msl == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_msl_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_st20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_st20_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcc == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_tcc_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcw == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_tcw_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_sm20_train[:,i,j].reshape(-1,1)], axis=1)                
                    if ecmwf_q == 1:
                        if q_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_q_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_q_train[:,q_col,i,j].reshape(-1,1)], axis=1)                
                #-----------------------------------------------------------------------------------------------    

                train_loc_p = train_seq_p.to_array().values.squeeze()[:, :, i, j]
                train_loc_quant = np.argmax(train_loc_p, axis=0)

                ### cut X_train due to N-past year values
                if (flag_past_year == 1) and (flag_use_w34 != 1):
                    if one_year_past == 1:
                        X_train = exog_set_train[(53-num_indent):]
                        y_train = train_loc_quant[53:]-1
                    if N_year_past == 1:
                        X_train = exog_set_train[((53-num_indent)+53*(N_years-1)):]
                        y_train = train_loc_quant[53*N_years:]-1
                elif (flag_use_w34 == 1) and (w34_pred == 1):
                    X_train = exog_set_train[-(w34_train.shape[0]):]
                    y_train = train_loc_quant[-(w34_train.shape[0]):]-1
                else:
                    X_train = exog_set_train[(53-num_indent):]
                    y_train = train_loc_quant[53:]-1

                X_train_all.extend(X_train)
                y_train_all.extend(y_train)

        X_train_all = np.array(X_train_all)
        y_train_all = np.array(y_train_all)

        print(X_train_all.shape)
        print(y_train_all.shape)
        X_train_all = np.nan_to_num(X_train_all)
        y_train_all = np.nan_to_num(y_train_all)

        ### classification
        if y_train_all.shape[0] != 0:
            print("!!!!!!!! Start classification >> "+name)

            if CLS == "RF":
                clf = RandomForestClassifier(n_estimators=nest, ccp_alpha=ccalpha, max_depth=max_depth, criterion=criterion)
            else:
                print("Not implemented!!")

            clf.fit(X_train_all, y_train_all)

            ### save the model to disk
            save_model_path = root_dir+"/model"
            if not os.path.exists(save_model_path):
                os.makedirs(save_model_path)
            filename = save_model_path+'/region_model_'+name+'.sav'
            pickle.dump(clf, open(filename, 'wb'))
                
            print("!!!!!!!! End classification >> "+name)

    print(">>>>>>>>>>>>>> Finish Overall <<<<<<<<<<<<<")

# # ------------------------------------------------------------

if flag_pred == 1:

    print("!!!!!!!! Start Predicting !!!!!!!!")

    for row in range(df_all_max.shape[0]):
        name = df_all_max.iloc[[row]].index.values[0]   
        print("!!!!", name)

        filename = root_dir+'/model/region_model_'+name+'.sav'
        clf = pickle.load(open(filename, 'rb'))
        d_name = root_dir+'/setting/setting_'+name+'.pkl'
        d_set = pickle.load(open(d_name, 'rb'))
    
        # -----------------
        flag_fct_bias = d_set["flag_fct_bias"]
        flag_biw = d_set["flag_biw"]
        flag_nino = d_set["flag_nino"]
        flag_solar = d_set["flag_solar"]
        flag_nao = d_set["flag_nao"]
        flag_pdo = d_set["flag_pdo"]
        flag_gl = d_set["flag_gl"]
        flag_qbo = d_set["flag_qbo"]
        flag_qbo_col = d_set["flag_qbo_col"]
        flag_rmm = d_set["flag_rmm"]
        flag_rmm_avg = d_set["flag_rmm_avg"]
        N_rmm_pst = d_set["N_rmm_pst"]
        flag_amp = d_set["flag_amp"]
        flag_another = d_set["flag_another"]
        flag_another_fct = d_set["flag_another_fct"]
        flag_use_w34 = d_set["flag_use_w34"]
        w34_pred = d_set["w34_pred"]
        w34_fct = d_set["w34_fct"]
        flag_use_ecmwf = d_set["flag_use_ecmwf"]
        flag_ecmwf_anom = d_set["flag_ecmwf_anom"]
        ecmwf_rsn = d_set["ecmwf_rsn"]
        ecmwf_sm100 = d_set["ecmwf_sm100"]
        ecmwf_sm20 = d_set["ecmwf_sm20"]
        ecmwf_msl = d_set["ecmwf_msl"]
        ecmwf_st20 = d_set["ecmwf_st20"]
        ecmwf_tcc = d_set["ecmwf_tcc"]
        ecmwf_tcw = d_set["ecmwf_tcw"]
        ecmwf_ci = d_set["ecmwf_ci"]
        ecmwf_sst = d_set["ecmwf_sst"]
        n_neig = d_set["n_neig"]
        sst_reduce = d_set["sst_reduce"]
        ecmwf_gh = d_set["ecmwf_gh"]
        gh_col = d_set["gh_col"]
        ecmwf_u = d_set["ecmwf_u"]
        u_col = d_set["u_col"]
        ecmwf_v = d_set["ecmwf_v"]
        v_col = d_set["v_col"]
        ecmwf_q = d_set["ecmwf_q"]
        q_col = d_set["q_col"]
        # -----------------

        loc_region = d_region[name]
        u1, d1, l1, r1 = loc_region[0]
        u2, d2, l2, r2 = u1, d1, l1, r1
        if len(loc_region) > 1:
            u2, d2, l2, r2 = loc_region[1]

        for idx in range(len(mask_flt[0])):

            if idx % 1000 == 0:
                print("Training", idx)

            i = mask_flt[0][idx]
            j = mask_flt[1][idx]  

            if ((i>=u1) and (i<d1) and (j>=l1) and (j<r1)) or ((i>=u2) and (i<d2) and (j>=l2) and (j<r2)):
                
                ### testing on trainingand validation set
                # ### Form training set
                window = list(np.arange(num_indent, num_indent+9))
                lower = train_terc[num_indent:,0,i,j]
                upper = train_terc[num_indent:,1,i,j]

                train_y_avg_rm_scale_mm = ((train_y_avg_rm_scale[:,i,j]-lower)/((upper-lower)+smallv))
                fct_train_mean_part = (fct_train_mean[:,i, j]-lower)/((upper-lower)+smallv)
                fct_train_std_part = (fct_train_std[:,i, j]-lower)/((upper-lower)+smallv)

                ### lag value
                pst_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], len(window)])
                for lgth_idx in range(len(window)):
                    rll = np.roll(train_y_avg_rm_scale_mm, window[lgth_idx])
                    pst_part[:, lgth_idx] = rll

                ### past N year value
                if flag_past_year == 1:
                    ## one year
                    if one_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], 1])
                        rll = np.roll(train_y_avg_rm_scale_mm, 53)
                        year_part[:, 0] = rll

                    ## two year
                    if N_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], N_years])
                        for N_year in range(N_years):
                            rll_N = np.roll(train_y_avg_rm_scale_mm, 53*(N_year+1))
                            year_part[:, N_year] = rll_N
                
                #-----------------------------------------------------------------------------------------------    
                ### Training data
                if flag_past_day == 1:
                    exog_set_train = np.concatenate([pst_part, 
                                                    fct_train_mean_part.reshape(-1,1), 
                                                    fct_train_std_part.reshape(-1,1)], axis=1)
                else:
                    exog_set_train = np.concatenate([fct_train_mean_part.reshape(-1,1), 
                                                        fct_train_std_part.reshape(-1,1)], axis=1)
                if flag_past_year == 1:
                    exog_set_train = np.concatenate([exog_set_train, year_part], axis=1)
                
                if flag_fct_bias == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        bias_mean_train[:, i, j].reshape(-1,1), 
                                                        bias_std_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_use_w34 == 1:
                    if w34_pred == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        w34_pred_train[:,:,i,j]], axis=1)                
                    if w34_fct == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            w34_fct_train[:,i,j].reshape(-1,1)], axis=1) 
                #-----------------------------------------------------------------------------------------------    
                if flag_biw[0] == 1:
                    exog_set_train = np.concatenate([exog_set_train, avg_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[1] == 1:
                    exog_set_train = np.concatenate([exog_set_train, std_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[2] == 1:
                    exog_set_train = np.concatenate([exog_set_train, skw_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[3] == 1:
                    exog_set_train = np.concatenate([exog_set_train, krt_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[4] == 1:
                    exog_set_train = np.concatenate([exog_set_train, med_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_nino == 1:                    
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        nin12_train.reshape(-1,1), nin34_train.reshape(-1,1)], axis=1)
                if flag_solar == 1:
                    exog_set_train = np.concatenate([exog_set_train, solar_train.reshape(-1,1)], axis=1)
                if flag_nao == 1:
                    exog_set_train = np.concatenate([exog_set_train, nao_train.reshape(-1,1)], axis=1)
                if flag_pdo == 1:
                    exog_set_train = np.concatenate([exog_set_train, pdo_train.reshape(-1,1)], axis=1)
                if flag_gl == 1:
                    exog_set_train = np.concatenate([exog_set_train, gl_train.reshape(-1,1)], axis=1)

                if flag_qbo == 1:
                    if flag_qbo_col == -1:
                        exog_set_train = np.concatenate([exog_set_train, qbo_train], axis=1)
                    else:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        qbo_train.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                if flag_rmm == 1:
                    if flag_rmm_avg == 1:
                        exog_set_train = np.concatenate([exog_set_train, rmm1_train.reshape(-1,1)], axis=1)
                        exog_set_train = np.concatenate([exog_set_train, rmm2_train.reshape(-1,1)], axis=1)
                    else:
                        exog_set_train = np.concatenate([exog_set_train, rmm1_train], axis=1)
                        exog_set_train = np.concatenate([exog_set_train, rmm2_train], axis=1)
                if flag_amp == 1:
                    exog_set_train = np.concatenate([exog_set_train, amp_train.reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_another == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        train_other[:, i, j].reshape(-1,1)], axis=1)
                if flag_another_fct == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        train_other_fct[:, i, j].reshape(-1,1)], axis=1)    

                #-----------------------------------------------------------------------------------------------    
                if flag_use_ecmwf == 1:
                    if ecmwf_rsn == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_rsn_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm100 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_sm100_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sst == 1:
                        ### k-nearest-neighbour method
                        if n_neig != 0:
                            tt = neigh.kneighbors([[i,j]], n_neig, return_distance=False)
                            sst_seq = ecmwf_sst_train[:, mask_sst[0][tt][0], mask_sst[1][tt][0]].squeeze()
                            if n_neig == 1:
                                sst_seq = sst_seq.reshape(-1,1)
                        ### correlation method
                        if sst_reduce in [0, 1]:
                            a = (sst_train-np.mean(sst_train, axis=0)).transpose([1,2,0])
                            b = (train_y_avg_rm_scale[:,i,j]-train_y_avg_rm_scale[:,i,j].mean()).reshape(-1,1)
                            sst_cor = np.dot(a, b).squeeze() / (np.std(sst_train, axis=0)*np.std(train_y_avg_rm_scale[:,i,j])*sst_train.shape[0])
                            sst_cor_loc = np.unravel_index(np.nanargmax(sst_cor), sst_cor.shape)
                            sst_seq = sst_train[:,sst_cor_loc[0], sst_cor_loc[1]].reshape(-1,1)
                        exog_set_train = np.concatenate([exog_set_train, sst_seq], axis=1)
                    if ecmwf_gh == 1:
                        if gh_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_gh_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_gh_train[:,gh_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_ci == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            ecmwf_ci_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_u == 1:
                        if u_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_u_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_u_train[:,u_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_v == 1:
                        if v_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_v_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_v_train[:,v_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_msl == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_msl_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_st20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_st20_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcc == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_tcc_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcw == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_tcw_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_sm20_train[:,i,j].reshape(-1,1)], axis=1)      
                    if ecmwf_q == 1:
                        if q_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_q_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                                ecmwf_q_train[:,q_col,i,j].reshape(-1,1)], axis=1)        
                #-----------------------------------------------------------------------------------------------   
                
                ### cut X_train due to N-past year values
                if (flag_past_year == 1) and (flag_use_w34 != 1):
                    if one_year_past == 1:
                        X_train = exog_set_train[(53-num_indent):]
                    if N_year_past == 1:
                        X_train = exog_set_train[((53-num_indent)+53*(N_years-1)):]
                elif (flag_use_w34 == 1) and (w34_pred == 1):
                    X_train = exog_set_train[-(w34_train.shape[0]):]
                else:
                    X_train = exog_set_train[(53-num_indent):]

                X_train = np.nan_to_num(X_train)
                y_train_prob = clf.predict_proba(X_train)
                save_train_prob_path = root_dir+"/y_train_prob_"+str(n)
                if not os.path.exists(save_train_prob_path):
                    os.makedirs(save_train_prob_path)
                np.save(save_train_prob_path+"/y_train_prob_"+str(i)+"by"+str(j)+"_"+str(n)+".npy", y_train_prob)


                # ########################################
                ### testing

                test_lower = test_terc[:,0,i,j]
                test_upper = test_terc[:,1,i,j]
                test_y_avg_rm_scale_mm = ((test_y_avg_rm_scale[:,i,j]-test_lower)/((test_upper-test_lower)+smallv))
                fct_test_mean_part = (fct_test_mean[:,i, j]-test_lower)/((test_upper-test_lower)+smallv)
                fct_test_std_part = (fct_test_std[:,i, j]-test_lower)/((test_upper-test_lower)+smallv)
                
                ### past days value
                roll_all = np.concatenate([train_y_avg_rm_scale_mm[-len(window):], test_y_avg_rm_scale_mm], axis=0)
                pst_test_part = np.zeros([roll_all.shape[0], len(window)])
                for lgth_idx in range(len(window)):
                    rll = np.roll(roll_all, window[lgth_idx])
                    pst_test_part[:, lgth_idx] = rll
                pst_test_part = pst_test_part[-len(test_y_avg_rm_scale_mm):]

                ### past years (same time) value
                if flag_past_year == 1:
                    if one_year_past == 1:
                        year_test_part = train_y_avg_rm_scale_mm[-53:].reshape(-1,1)
                    if N_year_past == 1:
                        year_test_part = np.zeros([53, N_years])
                        for N_year in range(N_years):
                            if N_year == 0:
                                year_test_part[:,0] = train_y_avg_rm_scale_mm[-53:]
                            else:
                                year_test_part[:,N_year] = train_y_avg_rm_scale_mm[-53*(N_year+1):-53*N_year]
                
                #-----------------------------------------------------------------------------------------------    
                ### Testing data
                if flag_past_day == 1:
                    exog_set_test = np.concatenate([pst_test_part, 
                                                    fct_test_mean_part.reshape(-1,1), 
                                                    fct_test_std_part.reshape(-1,1)], axis=1)
                else:
                    exog_set_test = np.concatenate([fct_test_mean_part.reshape(-1,1), 
                                                    fct_test_std_part.reshape(-1,1)], axis=1)
                if flag_past_year == 1:
                    exog_set_test = np.concatenate([exog_set_test, year_test_part], axis=1)
                
                if flag_fct_bias == 1:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    bias_mean_test[:, i, j].reshape(-1,1), 
                                                    bias_std_test[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_use_w34 == 1:
                    if w34_pred == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        w34_pred_test[:,:,i,j]], axis=1)                
                    if w34_fct == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                            w34_fct_test[:,i,j].reshape(-1,1)], axis=1) 
                #-----------------------------------------------------------------------------------------------    
                if flag_biw[0] == 1:
                    exog_set_test = np.concatenate([exog_set_test, avg_test[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[1] == 1:
                    exog_set_test = np.concatenate([exog_set_test, std_test[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[2] == 1:
                    exog_set_test = np.concatenate([exog_set_test, skw_test[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[3] == 1:
                    exog_set_test = np.concatenate([exog_set_test, krt_test[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[4] == 1:
                    exog_set_test = np.concatenate([exog_set_test, med_test[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_nino == 1:                    
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    nin12_test.reshape(-1,1), 
                                                    nin34_test.reshape(-1,1)], axis=1)
                if flag_solar == 1:
                    exog_set_test = np.concatenate([exog_set_test, solar_test.reshape(-1,1)], axis=1)
                if flag_nao == 1:
                    exog_set_test = np.concatenate([exog_set_test, nao_test.reshape(-1,1)], axis=1)
                if flag_pdo == 1:
                    exog_set_test = np.concatenate([exog_set_test, pdo_test.reshape(-1,1)], axis=1)
                if flag_gl == 1:
                    exog_set_test = np.concatenate([exog_set_test, gl_test.reshape(-1,1)], axis=1)

                if flag_qbo == 1:
                    if flag_qbo_col == -1:
                        exog_set_test = np.concatenate([exog_set_test, qbo_test], axis=1)
                    else:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        qbo_test.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    

                if flag_rmm == 1:
                    if flag_rmm_avg == 1:
                        exog_set_test = np.concatenate([exog_set_test, rmm1_test.reshape(-1,1)], axis=1)
                        exog_set_test = np.concatenate([exog_set_test, rmm2_test.reshape(-1,1)], axis=1)
                    else:
                        exog_set_test = np.concatenate([exog_set_test, rmm1_test], axis=1)
                        exog_set_test = np.concatenate([exog_set_test, rmm2_test], axis=1)

                if flag_amp == 1:
                    exog_set_test = np.concatenate([exog_set_test, amp_test.reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    

                if flag_another == 1:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    test_other[:, i, j].reshape(-1,1)], axis=1)
                if flag_another_fct == 1:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    test_other_fct[:, i, j].reshape(-1,1)], axis=1)    

                #-----------------------------------------------------------------------------------------------    
                if flag_use_ecmwf == 1:
                    if ecmwf_rsn == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_rsn_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm100 == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_sm100_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sst == 1:
                        ### k-nearest-neighbour method
                        if n_neig != 0:
                            sst_seq_test = ecmwf_sst_test[:, mask_sst[0][tt][0], mask_sst[1][tt][0]]
                            if n_neig == 1:
                                sst_seq_test = sst_seq_test.reshape(-1,1)
                        ### correlation method
                        if sst_reduce in [0, 1]:
                            sst_seq_test = sst_test[:,sst_cor_loc[0], sst_cor_loc[1]].reshape(-1,1)
                        exog_set_test = np.concatenate([exog_set_test, sst_seq_test], axis=1)
                    if ecmwf_gh == 1:
                        if gh_col == -1:
                            exog_set_test = np.concatenate([exog_set_test, ecmwf_gh_test[:,:,i,j]], axis=1)
                        else:
                            exog_set_test = np.concatenate([exog_set_test, 
                                                            ecmwf_gh_test[:,gh_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_ci == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_ci_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_u == 1:
                        if u_col == -1:
                            exog_set_test = np.concatenate([exog_set_test, ecmwf_u_test[:,:,i,j]], axis=1)
                        else:
                            exog_set_test = np.concatenate([exog_set_test, 
                                                            ecmwf_u_test[:,u_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_v == 1:
                        if v_col == -1:
                            exog_set_test = np.concatenate([exog_set_test, ecmwf_v_test[:,:,i,j]], axis=1)
                        else:
                            exog_set_test = np.concatenate([exog_set_test, 
                                                            ecmwf_v_test[:,v_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_msl == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_msl_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_st20 == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_st20_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcc == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_tcc_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcw == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_tcw_test[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm20 == 1:
                        exog_set_test = np.concatenate([exog_set_test, 
                                                        ecmwf_sm20_test[:,i,j].reshape(-1,1)], axis=1)      
                    if ecmwf_q == 1:
                        if q_col == -1:
                            exog_set_test = np.concatenate([exog_set_test, ecmwf_q_test[:,:,i,j]], axis=1)
                        else:
                            exog_set_test = np.concatenate([exog_set_test, 
                                                            ecmwf_q_test[:,q_col,i,j].reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                
                ### form X_test
                X_test = exog_set_test
                X_test = np.nan_to_num(X_test)
                y_test_prob = clf.predict_proba(X_test)
                save_seq_prob_path = root_dir+"/y_test_prob_"+str(n)
                if not os.path.exists(save_seq_prob_path):
                    os.makedirs(save_seq_prob_path)
                np.save(save_seq_prob_path+"/y_test_prob_"+str(i)+"by"+str(j)+"_"+str(n)+".npy", y_test_prob)

                # ------------------------------------------------------------
                ### testing on 2020
                window = list(np.arange(num_indent, num_indent+9))
                lower, upper = train_terc[num_indent:,0,i,j], train_terc[num_indent:,1,i,j]
                test_lower, test_upper = test_terc[:,0,i,j], test_terc[:,1,i,j]

                train_y_avg_rm_scale_mm = ((train_y_avg_rm_scale[:,i,j]-lower)/((upper-lower)+smallv))
                test_y_avg_rm_scale_mm = ((test_y_avg_rm_scale[:,i,j]-test_lower)/((test_upper-test_lower)+smallv))
                train_y_avg_rm_scale_mm = np.concatenate((train_y_avg_rm_scale_mm, test_y_avg_rm_scale_mm))

                test_2020_lower = test_2020_terc[:,0,i,j]
                test_2020_upper = test_2020_terc[:,1,i,j]
                test_2020_y_avg_rm_scale_mm = ((test_2020_y_avg_rm_scale[:,i,j]-test_2020_lower)/((test_2020_upper-test_2020_lower)+smallv))
                fct_2020_mean_part = (fct_2020_mean[:,i, j]-test_2020_lower)/((test_2020_upper-test_2020_lower)+smallv)
                fct_2020_std_part = (fct_2020_std[:,i, j]-test_2020_lower)/((test_2020_upper-test_2020_lower)+smallv)
                
                ### past days value
                roll_all = np.concatenate([train_y_avg_rm_scale_mm[-len(window):], test_2020_y_avg_rm_scale_mm], axis=0)
                pst_2020_part = np.zeros([roll_all.shape[0], len(window)])
                for lgth_idx in range(len(window)):
                    rll = np.roll(roll_all, window[lgth_idx])
                    pst_2020_part[:, lgth_idx] = rll
                pst_2020_part = pst_2020_part[-len(test_2020_y_avg_rm_scale_mm):]

                ### past years (same time) value
                if flag_past_year == 1:
                    if one_year_past == 1:
                        year_2020_part = train_y_avg_rm_scale_mm[-53:].reshape(-1,1)
                    if N_year_past == 1:
                        year_2020_part = np.zeros([53, N_years])
                        for N_year in range(N_years):
                            if N_year == 0:
                                year_2020_part[:,0] = train_y_avg_rm_scale_mm[-53:]
                            else:
                                year_2020_part[:,N_year] = train_y_avg_rm_scale_mm[-53*(N_year+1):-53*N_year]
                
                #-----------------------------------------------------------------------------------------------    
                ### Testing data
                if flag_past_day == 1:
                    exog_set_2020 = np.concatenate([pst_2020_part, 
                                                    fct_2020_mean_part.reshape(-1,1), 
                                                    fct_2020_std_part.reshape(-1,1)], axis=1)
                else:
                    exog_set_2020 = np.concatenate([fct_2020_mean_part.reshape(-1,1), 
                                                    fct_2020_std_part.reshape(-1,1)], axis=1)
                if flag_past_year == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, year_2020_part], axis=1)
                
                if flag_fct_bias == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, 
                                                    bias_mean_2020[:, i, j].reshape(-1,1), 
                                                    bias_std_2020[:, i, j].reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                ## TODO: w56 use w34 fct and pred
                if flag_use_w34 == 1:
                    if w34_pred == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            w34_pred_test[:,:,i,j]], axis=1)                
                    if w34_fct == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            w34_fct_test[:,i,j].reshape(-1,1)], axis=1) 
                #-----------------------------------------------------------------------------------------------    
                if flag_biw[0] == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, avg_2020[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[1] == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, std_2020[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[2] == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, skw_2020[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[3] == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, krt_2020[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[4] == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, med_2020[:, i, j].reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                if flag_nino == 1:                    
                    exog_set_2020 = np.concatenate([exog_set_2020, 
                                                    nin12_2020.reshape(-1,1), 
                                                    nin34_2020.reshape(-1,1)], axis=1)
                if flag_solar == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, solar_2020.reshape(-1,1)], axis=1)
                if flag_nao == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, nao_2020.reshape(-1,1)], axis=1)
                if flag_pdo == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, pdo_2020.reshape(-1,1)], axis=1)
                if flag_gl == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, gl_2020.reshape(-1,1)], axis=1)

                if flag_qbo == 1:
                    if flag_qbo_col == -1:
                        exog_set_2020 = np.concatenate([exog_set_2020, qbo_2020], axis=1)
                    else:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        qbo_2020.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                if flag_rmm == 1:
                    if flag_rmm_avg == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, rmm1_2020.reshape(-1,1)], axis=1)
                        exog_set_2020 = np.concatenate([exog_set_2020, rmm2_2020.reshape(-1,1)], axis=1)
                    else:
                        exog_set_2020 = np.concatenate([exog_set_2020, rmm1_2020], axis=1)
                        exog_set_2020 = np.concatenate([exog_set_2020, rmm2_2020], axis=1)

                if flag_amp == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, amp_2020.reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    

                if flag_another == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, 
                                                    test_2020_other[:, i, j].reshape(-1,1)], axis=1)
                if flag_another_fct == 1:
                    exog_set_2020 = np.concatenate([exog_set_2020, 
                                                    test_2020_other_fct[:, i, j].reshape(-1,1)], axis=1)    
                #-----------------------------------------------------------------------------------------------    
                if flag_use_ecmwf == 1:
                    if ecmwf_rsn == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_rsn_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm100 == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_sm100_2020[:,i,j].reshape(-1,1)], axis=1)
                    ### TODO: SST training and testing match problem
                    if ecmwf_sst == 1:
                        ### k-nearest-neighbour method
                        if n_neig != 0:
                            tt = neigh.kneighbors([[i,j]], n_neig, return_distance=False)
                            sst_seq_2020 = ecmwf_sst_2020[:, mask_sst[0][tt][0], mask_sst[1][tt][0]]
                            if n_neig == 1:
                                sst_seq_2020 = sst_seq_2020.reshape(-1,1)
                        ### correlation method
                        if sst_reduce in [0, 1]:
                            sst_seq_2020 = sst_test[:,sst_cor_loc[0], sst_cor_loc[1]].reshape(-1,1)
                        exog_set_2020 = np.concatenate([exog_set_2020, sst_seq_2020], axis=1)
                    if ecmwf_gh == 1:
                        if gh_col == -1:
                            exog_set_2020 = np.concatenate([exog_set_2020, ecmwf_gh_2020[:,:,i,j]], axis=1)
                        else:
                            exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            ecmwf_gh_2020[:,gh_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_ci == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_ci_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_u == 1:
                        if u_col == -1:
                            exog_set_2020 = np.concatenate([exog_set_2020, ecmwf_u_2020[:,:,i,j]], axis=1)
                        else:
                            exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            ecmwf_u_2020[:,u_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_v == 1:
                        if v_col == -1:
                            exog_set_2020 = np.concatenate([exog_set_2020, ecmwf_v_2020[:,:,i,j]], axis=1)
                        else:
                            exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            ecmwf_v_2020[:,v_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_msl == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_msl_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_st20 == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_st20_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcc == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_tcc_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcw == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_tcw_2020[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm20 == 1:
                        exog_set_2020 = np.concatenate([exog_set_2020, 
                                                        ecmwf_sm20_2020[:,i,j].reshape(-1,1)], axis=1)      
                    if ecmwf_q == 1:
                        if q_col == -1:
                            exog_set_2020 = np.concatenate([exog_set_2020, ecmwf_q_2020[:,:,i,j]], axis=1)
                        else:
                            exog_set_2020 = np.concatenate([exog_set_2020, 
                                                            ecmwf_q_2020[:,q_col,i,j].reshape(-1,1)], axis=1)
                #-----------------------------------------------------------------------------------------------    
                ### form X_test
                X_2020 = exog_set_2020
                X_2020 = np.nan_to_num(X_2020)

                y_2020_prob = clf.predict_proba(X_2020)
                save_seq_prob_path = root_dir+"/y_2020_prob_"+str(n)
                if not os.path.exists(save_seq_prob_path):
                    os.makedirs(save_seq_prob_path)
                np.save(save_seq_prob_path+"/y_2020_prob_"+str(i)+"by"+str(j)+"_"+str(n)+".npy", y_2020_prob)

print("!!!!!!!! End Predicting !!!!!!!!")
print(root_dir)

