########################################################################
# Script for generation temperature predicton for each location

# To run the script:
#   python3 main_t2m.py argv1 argv2
#   @param argv1: Classification method -- [RF, MLP. KNN]
#                 Submission: RF
#   @param argv2: "w34" for week 3-4 prediction
#                 "w56" for week 5-6 prediction
######################################################################

import xarray as xr
import numpy as np
import pandas as pd
import sys, pickle, os
from sklearn.preprocessing import MinMaxScaler
from sklearn.ensemble import RandomForestClassifier
from sklearn.neighbors import NearestNeighbors
from skimage.measure import block_reduce

from sklearn.neural_network import MLPClassifier
from sklearn.linear_model import LogisticRegression
from sklearn.neighbors import KNeighborsClassifier
from sklearn import tree
import matplotlib.pyplot as plt
from xgboost import XGBClassifier
import time
# ---------------------------------------------------------------------

cache_path = "../../data/newdata"
ecmwf_path = "../download_data/data"
root_path = "../"

time_train_start, time_train_end = '2000', str(int(sys.argv[34])-1)      # train
time_valid_start, time_valid_end = sys.argv[34], sys.argv[35]            # valid
# time_train_start, time_train_end = '2000', '2017' # train
# time_valid_start, time_valid_end = '2018', '2019' # valid
time_test = '2020'                            # test

# ---------------------------------------------------------------------
if v == 't2m':   
    from scripts_t2m import *
else:
    from scripts_tp import *
    
### argument
CLS = sys.argv[1]                           ### Choose classification method. Suggestion: RF
wk = sys.argv[2]                            ### Choose 


### run week 3-4 prediction
if wk == "w34":

    flag_w34 = 1                            ### predict w34
    flag_w56 = 0
    
    num_indent = 4

    ### RF hyper parameters
    nest = 100
    ccalpha = 0
    max_depth = 10                          ### 10 for w34, 1 for w56
    criterion = "gini"                      ### entropy for w34, gini for w56

    ### feature parametrs
    flag_past_day = 1
    flag_past_year = 10                     ### whether to use past years' same time value
    one_year_past = 0                       ### if flag_past_year == 1, choose whether only use past one year
    N_year_past = 1-one_year_past           ### if flag_past_year == 1, choose whether only use past N years
    N_years = 1

    flag_fct_bias = int(sys.argv[3])

    flag_biw = [int(sys.argv[4]), int(sys.argv[5]), int(sys.argv[6]), int(sys.argv[7]), int(sys.argv[8])]                   ### whether to use biweekly statistics, [avg, std, skew, kurt, median]

    flag_nino = int(sys.argv[9])
    flag_solar = int(sys.argv[10])                         ### Sunspot Cycles [0: not use; 1: normal; 2: smoothed]
    flag_nao = int(sys.argv[11])                           ### NAO data (North Atlantic Oscillation) [0: not use; 1: use]
    flag_pdo = int(sys.argv[12])                           ### PDO data (Pacific Decadal Oscillation) [0: not use; 1: use]
    flag_gl = int(sys.argv[13])                            ### Great Lakes Ice Cover [0: not use; 1: use]

    flag_qbo = int(sys.argv[14])
    flag_qbo_col = int(sys.argv[29])      # [-1, ..., 6]

    flag_another = int(sys.argv[15])                       ### another variable (tp or t2m)
    flag_another_fct = int(sys.argv[16])

    flag_use_w34 = 0
    w34_pred = 0
    w34_fct = 1
    
    # ---------
    flag_use_ecmwf = 1
    flag_ecmwf_anom = 1

    ecmwf_rsn = int(sys.argv[17])
    ecmwf_sm100 = int(sys.argv[18])
    ecmwf_sm20 = int(sys.argv[19])
    ecmwf_msl = int(sys.argv[20])
    ecmwf_st20 = int(sys.argv[21])
    ecmwf_tcc = int(sys.argv[22])
    ecmwf_tcw = int(sys.argv[23])
    ecmwf_ci = int(sys.argv[24])

    ecmwf_gh = int(sys.argv[25])
    gh_col = int(sys.argv[30])      # [-1, ..., 9]
    ecmwf_u = int(sys.argv[26])
    u_col = int(sys.argv[31])
    ecmwf_v = int(sys.argv[27])
    v_col = int(sys.argv[32])
    ecmwf_q = int(sys.argv[28])
    q_col = int(sys.argv[33])       # [-1, ..., 6]

### run week 5-6 prediction
elif wk == "w56":

    flag_w56 = 1                            ### predict w56
    flag_w34 = 0

    num_indent = 6

    ### RF hyper parameters
    nest = 100
    ccalpha = 0
    max_depth = 10                               ### 10 for w34, 1 for w56
    criterion = "entropy"                    ### entropy for w34, gini for w56
    
    ### feature parametrs
    flag_past_day = 1
    flag_past_year = 1                      ### whether to use past years' same time value
    one_year_past = 0                       ### if flag_past_year == 1, choose whether only use past one year
    N_year_past = 1-one_year_past           ### if flag_past_year == 1, choose whether only use past N years
    N_years = 1

    flag_fct_bias = int(sys.argv[3])

    flag_biw = [int(sys.argv[4]), int(sys.argv[5]), int(sys.argv[6]), int(sys.argv[7]), int(sys.argv[8])]  ### whether to use biweekly statistics, [avg, std, skew, kurt, median]

    flag_nino = int(sys.argv[9])
    flag_solar = int(sys.argv[10])                         ### Sunspot Cycles [0: not use; 1: normal; 2: smoothed]
    flag_nao = int(sys.argv[11])                           ### NAO data (North Atlantic Oscillation) [0: not use; 1: use]
    flag_pdo = int(sys.argv[12])                           ### PDO data (Pacific Decadal Oscillation) [0: not use; 1: use]
    flag_gl = int(sys.argv[13])                            ### Great Lakes Ice Cover [0: not use; 1: use]

    flag_qbo = int(sys.argv[14])
    flag_qbo_col = int(sys.argv[29])      # [-1, ..., 6]

    flag_another = int(sys.argv[15])                       ### another variable (tp or t2m)
    flag_another_fct = int(sys.argv[16])

    flag_use_w34 = 0
    w34_pred = 0
    w34_fct = 1

    flag_use_ecmwf = 1
    flag_ecmwf_anom = 1

    ecmwf_rsn = int(sys.argv[17])
    ecmwf_sm100 = int(sys.argv[18])
    ecmwf_sm20 = int(sys.argv[19])
    ecmwf_msl = int(sys.argv[20])
    ecmwf_st20 = int(sys.argv[21])
    ecmwf_tcc = int(sys.argv[22])
    ecmwf_tcw = int(sys.argv[23])
    ecmwf_ci = int(sys.argv[24])

    ecmwf_gh = int(sys.argv[25])
    gh_col = int(sys.argv[30])      # [-1, ..., 9]
    ecmwf_u = int(sys.argv[26])
    u_col = int(sys.argv[31])
    ecmwf_v = int(sys.argv[27])
    v_col = int(sys.argv[32])
    ecmwf_q = int(sys.argv[28])
    q_col = int(sys.argv[33])       # [-1, ..., 6]
    
      
else:
    print("Wrong week index!!!")

# ---------------
### no need to change
flag_preprocess = 1
flag_pred = 1

scale = False
n = 10
smallv = 1e-10

# ---------------------------------------------------------------------
### result saved path

root_dir = (f"{root_path}/result_sys_valid_tp_ecmwf/unit1/cv_{time_valid_start}_{time_valid_end}/RegModel_RF_{criterion}_w34{flag_w34}_w56{flag_w56}_"
            f"fctB{flag_fct_bias}_biw{flag_biw[0]}{flag_biw[1]}{flag_biw[2]}{flag_biw[3]}{flag_biw[4]}_"
            f"nino{flag_nino}_solar{flag_solar}_nao{flag_nao}_pdo{flag_pdo}_gl{flag_gl}_"
            f"qbo{flag_qbo}{flag_qbo_col}_oth{flag_another}{flag_another_fct}_"
            f"ecmwf{flag_use_ecmwf}{flag_ecmwf_anom}_"
          f"rsn{ecmwf_rsn}_sm100{ecmwf_sm100}_msl{ecmwf_msl}_st20{ecmwf_st20}_tcc{ecmwf_tcc}_tcw{ecmwf_tcw}_sm20{ecmwf_sm20}_ci{ecmwf_ci}_"
            f"gh{ecmwf_gh}{gh_col}u{ecmwf_u}{u_col}v{ecmwf_v}{v_col}q{ecmwf_q}{q_col}"
            )

if not os.path.exists(root_dir):
    os.makedirs(root_dir)
print("\n>>>>>>>>>", root_dir, "<<<<<<<<<<<\n")

# --------------------------------------------
### train output
try:
    obs_2000_2019 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc")
except:
    obs_2000_2019 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr")
### test output
try:
    obs_2020 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc")
except:
    obs_2020 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr")
### train input
try:
    hind_2000_2019 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc")
except:
    hind_2000_2019 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr")
### test input
try:
    fct_2020 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc")
except:
    fct_2020 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr")

### tercile
tercile_file = f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.nc"
tercile_edges = xr.open_dataset(tercile_file)

if flag_w34 == 1:
    obs_2000_2019_w34 = obs_2000_2019.isel(lead_time=[0])
    obs_2020_w34 = obs_2020.isel(lead_time=[0])

    hind_2000_2019_w34 = hind_2000_2019.isel(lead_time=[0])
    fct_2020_w34 = fct_2020.isel(lead_time=[0])
    
    tercile_edges_v = tercile_edges.isel(lead_time=0)[v]

elif flag_w56 == 1:
    obs_2000_2019_w34 = obs_2000_2019.isel(lead_time=[1])
    obs_2020_w34 = obs_2020.isel(lead_time=[1])

    hind_2000_2019_w34 = hind_2000_2019.isel(lead_time=[1])
    fct_2020_w34 = fct_2020.isel(lead_time=[1])

    tercile_edges_v = tercile_edges.isel(lead_time=1)[v]

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = mask.notnull()[v]

### Mask training input
hind_2000_2019_w34_msk = hind_2000_2019_w34.where(obs_2020_w34.isel(forecast_time=0, lead_time=0, drop=True).notnull())

# --------------------------------------------
### averaged removed sequences
train_data = obs_2000_2019_w34.sel(forecast_time=slice(time_train_start, time_train_end))[v].values.transpose(1,0,2,3)
train_data = np.mean(train_data, axis=1)

test_data = obs_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.transpose(1,0,2,3)
test_data = np.mean(test_data, axis=1)

# if v == 'tp':
#     train_data, Orig_1Dplot, After_1Dplot = Fill_Nan_pastInfor(train_data)

train_y = train_data
test_y = test_data

train_avg_rm = train_data[num_indent:, :, :]
test_avg_rm = test_data

train_y_avg_rm = train_avg_rm
test_y_avg_rm = test_avg_rm

if scale:
    print(">>>>>>>> SCALE Sequence")
    scaler = MinMaxScaler()
    train_y_avg_rm_scale = scaler.fit_transform(train_y_avg_rm.reshape(train_y_avg_rm.shape[0], -1)).reshape(train_y_avg_rm.shape)
    test_y_avg_rm_scale = scaler.transform(test_y_avg_rm.reshape(test_y_avg_rm.shape[0], -1)).reshape(test_y_avg_rm.shape)
else:
    print(">>>>>>>> NOT SCALE Sequence")
    train_y_avg_rm_scale = train_y_avg_rm
    test_y_avg_rm_scale = test_y_avg_rm

if v == "t2m":
    mask_flt = np.where(~np.isnan(train_avg_rm[316]))
elif v == "tp":
    mask_flt = np.where(~np.isnan(train_data[0]))
    train_y_avg_rm_scale = np.nan_to_num(train_y_avg_rm_scale)
    test_y_avg_rm_scale = np.nan_to_num(test_y_avg_rm_scale)

# -------------------------------------------

fct_data = hind_2000_2019_w34.sel(forecast_time=slice(time_train_start, time_train_end))[v].values.squeeze().transpose(1,0,2,3)
if v == "tp":
    fct_data = np.nan_to_num(fct_data)
fct_data_mean = np.mean(fct_data, axis=1)
fct_data_std = np.std(fct_data, axis=1)

fct_test = hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.squeeze().transpose(1,0,2,3)
if v == "tp":
    fct_test = np.nan_to_num(fct_test)
fct_test_mean = np.mean(fct_test, axis=1)
fct_test_std = np.std(fct_test, axis=1)

fct_train_mean = fct_data_mean[num_indent:, :, :]
fct_train_std = fct_data_std[num_indent:, :, :]

# -------------------------------------------
if flag_fct_bias == 1:

    obs_2000_2019_w34_add = add_year_week_coords(obs_2000_2019_w34)
    hind_2000_2019_w34_add = add_year_week_coords(hind_2000_2019_w34)
    
    hind_train_add = hind_2000_2019_w34_add.sel(forecast_time=slice(time_train_start,time_train_end))
    hind_test_add = hind_2000_2019_w34_add.sel(forecast_time=slice(time_valid_start, time_valid_end))
    obs_train_add = obs_2000_2019_w34_add.sel(forecast_time=slice(time_train_start,time_train_end))
    obs_test_add = obs_2000_2019_w34_add.sel(forecast_time=slice(time_valid_start, time_valid_end))

    bias_mean_2000_2018 = (hind_train_add.mean('realization') - obs_train_add).groupby('week').mean().compute()
    bias_std_2000_2018 = (hind_train_add.mean('realization') - obs_train_add).groupby('week').std().compute()

    bias_mean_train = bias_mean_2000_2018.sel(week=hind_train_add.week)[v].squeeze().values
    bias_std_train = bias_std_2000_2018.sel(week=hind_train_add.week)[v].squeeze().values

    bias_mean_test = bias_mean_2000_2018.sel(week=hind_test_add.week)[v].squeeze().values
    bias_std_test = bias_std_2000_2018.sel(week=hind_test_add.week)[v].squeeze().values

    bias_mean_train = bias_mean_train[num_indent:, :, :]
    bias_std_train = bias_std_train[num_indent:, :, :]

# ---------------------------------------------------------------------

def form_ds(pred_test, time, ds_cmpr=hind_2000_2019_w34, tercile_edges=tercile_edges_v, mask=mask_v):

    pred_test = np.expand_dims(pred_test, axis=1)

    ds_cmpr = ds_cmpr.sel(forecast_time=time)
    pred_test_da = xr.DataArray(pred_test,
                            dims=['forecast_time', 'lead_time', 'latitude', 'longitude'],
                            coords={'forecast_time': ds_cmpr.forecast_time,
                                    'lead_time': ds_cmpr.lead_time,
                                    'latitude': ds_cmpr.latitude,
                                    'longitude': ds_cmpr.longitude},
                                    )

    preds_test_add = add_valid_time_from_forecast_reference_time_and_lead_time(pred_test_da)
    preds_test_add = preds_test_add.to_dataset(name=v)
    
    preds_test_p = make_probabilistic(preds_test_add.expand_dims('realization'), tercile_edges, mask=mask)

    # -----
    ds = add_year_week_coords(preds_test_add.expand_dims('realization'))
    groupby_coord = 'week'
    tercile_edges_v_1 = tercile_edges.sel({groupby_coord: ds.coords[groupby_coord]}).values

    return preds_test_p, tercile_edges_v_1

### -----
def read_monthly(name, data_flag='nino'):

    if data_flag == 'nino':
        path = f'{root_path}/data/nino/'
        data =  pd.read_csv(path+name, sep="\n")
        data = data[:151]
        foo = lambda x: pd.Series([i for i in x.split('   ')])

    if data_flag == 'oscil':
        path = f'{root_path}/data/Oscillation/'
        data =  pd.read_csv(path+name, sep="\n", header=None)
        data = data[:-1]
        foo = lambda x: pd.Series([i for i in x.split('  ')])

    data = data.iloc[:,0].apply(foo)
    data.columns = [str(i) for i in range(13)]
    data = data.set_index('0')
    data_cut = data.iloc[-21:, :]

    df = pd.DataFrame()
    n_4 = [2,3,5,6,8,9,11]
    n_5 = [1,4,7,10,12]

    for col in data_cut.columns:
        if int(col) in n_4:
            p4 = pd.concat([data_cut[col]] * (4), axis=1, ignore_index=True)
            p4.columns = [col+"_1", col+"_2", col+"_3", col+"_4"]
            df = pd.concat([df, p4], axis=1)
        elif int(col) in n_5:
            p5 = pd.concat([data_cut[col]] * (5), axis=1, ignore_index=True)
            p5.columns = [col+"_1", col+"_2", col+"_3", col+"_4", col+"_5"]
            df = pd.concat([df, p5], axis=1)
    
    df_data = df[:21]
    df_data_list = []

    for i in range(df_data.shape[0]):
        df_data_list.extend(df_data.iloc[i, :])
    df_data_list = np.array(df_data_list)

    df_data_list = df_data_list[:-num_indent]

    df_train = df_data_list[:len(train_avg_rm)]
    df_test = df_data_list[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]

    df_train = df_train.astype(float)
    df_test = df_test.astype(float)

    return df_train, df_test


### -----

def read_daily(df):

    df_lst = []
    dates = xr.cftime_range(start='20000102',freq='7D', periods=53)
    for row_idx in range(df.shape[0]):
        row = df.iloc[row_idx, :]
        mon, day = int(row["MM"]), int(row["DD"])
        
        for i, f in enumerate(dates):
            if (mon == f.month) and (day == f.day):
                tmp = df.iloc[row_idx:row_idx+14, -1].mean()
                df_lst.append(tmp)
            
    df_lst = np.array(df_lst)
    df_lst = df_lst[:-num_indent]

    df_train = df_lst[:len(train_avg_rm)]
    df_test = df_lst[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]
    
    return df_train, df_test

### -----

def biw_varb(path):
    root_dir = f"{root_path}/data/biw_stat_"+v+"_"+wk+"/"
    data = np.load(root_dir+path)
    data_all = data

    for t in range(20):
        data_all = np.concatenate([data_all, data], axis=0)
    data_all = data_all[num_indent:, :, :]

    data_train = data_all[:len(train_avg_rm), :, :]
    data_test = data_all[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm), :, :]

    return data_train, data_test

# --------------------------------------------------------------

train_seq_p, train_terc = form_ds(train_y, slice(time_train_start,time_train_end), 
                                    ds_cmpr=hind_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end)))
test_seq_p, test_terc = form_ds(test_y, slice(time_valid_start, time_valid_end), ds_cmpr=hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end)))

# -------
### El NINO
if flag_nino == 1:
    nin12_train, nin12_test = read_monthly("nino12.long.anom.data")
    nin34_train, nin34_test = read_monthly("nino34.long.anom.data")

# -------
### Oscillation
if flag_nao == 1:
    nao_train, nao_test = read_monthly("norm.nao.monthly.b5001.current.ascii.table.txt", data_flag='oscil')
if flag_pdo == 1:
    pdo_train, pdo_test = read_monthly("ersst.v5.pdo.dat.txt", data_flag='oscil')

# -------
### biweekly statistics
avg_train, avg_test = biw_varb("biweekly_historical_avgs.npy")
std_train, std_test = biw_varb("biweekly_historical_std.npy")
med_train, med_test = biw_varb("biweekly_historical_med.npy")
skw_train, skw_test = biw_varb("biweekly_historical_skew.npy")
krt_train, krt_test = biw_varb("biweekly_historical_kurt.npy")

# -------
### Sunspot Cycles
if flag_solar != 0:

    n_4 = [2,3,5,6,8,9,11]
    n_5 = [1,4,7,10,12]

    solar = pd.read_json(f"{root_path}/data/observed-solar-cycle-indices.json")
    solar = solar[solar['time-tag'] >= '2000-01']
    solar = solar[solar['time-tag'] <= '2020-12']
    solar_ssn = solar[["time-tag", "ssn", "smoothed_ssn"]]

    solar_list = []
    for rol in range(len(solar_ssn)):
        ssn = solar_ssn.iloc[rol, flag_solar]
        mm = int(solar_ssn.iloc[rol, 0].split('-')[-1])

        if mm in n_4:
            solar_list.extend([ssn]*4)
        elif mm in n_5:
            solar_list.extend([ssn]*5)

    solar_list = np.array(solar_list)
    solar_list = solar_list[:-num_indent]

    solar_train = solar_list[:len(train_avg_rm)]
    solar_test = solar_list[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm)]

# -------
### Great Lakes Ice Cover
if flag_gl == 1:
    path = f'{root_path}/data/great_lake/N_seaice_extent_daily_v3.0.csv'
    data_gl =  pd.read_csv(path, header=1)
    data_gl = data_gl.iloc[: , :-2]
    data_gl.columns = ['YYYY', 'MM', 'DD', '10^6 sq km']
    data_gl = data_gl[data_gl["YYYY"]>=2000]
    data_gl = data_gl[data_gl["YYYY"]<=2020]

    gl_train, gl_test = read_daily(data_gl)

# -------
### QBO
if flag_qbo == 1:
    n_4 = [2,3,5,6,8,9,11]
    n_5 = [1,4,7,10,12]

    path = f"{root_path}/data/Oscillation/qbo.dat"
    pdo = pd.read_csv(path, sep="\n", header=None)
    pdo = pdo[7+(12*(2000-1953)):-11]
    foo = lambda x: pd.Series([t for t in [i for i in x.split(' ')] if t])
    pdo = pdo.iloc[:,0].apply(foo).iloc[:,1:]
    pdo.columns = ["YYMM", "70hPaN", "50hPaN", "40hPaN", "30hPaN", "20hPaN", "15hPaN", "10hPaN"]
    pdo[['YY','MM']] = pdo.YYMM.apply(lambda x: pd.Series([x[index : index+2] for index in range(0, len(x), 2)]))
    pdo = pdo.iloc[:,1:]

    a = [i+1 for i in range(12)]
    rep = [4 if i in n_4 else 5 for i in a]*21
    pdo["times"] = rep

    pdo_repeat = pdo.loc[pdo.index.repeat(pdo.times)]
    pdo_repeat = pdo_repeat.iloc[:-num_indent, :]
    qbo_train = pdo_repeat.iloc[:len(train_avg_rm), :-3]
    qbo_test = pdo_repeat.iloc[len(train_avg_rm):len(train_avg_rm)+len(test_avg_rm), :-3]
    qbo_train, qbo_test = qbo_train.astype(float), qbo_test.astype(float)

# -------
### Another variable (tp or t2m)
if v == "tp": 
    v_other = "t2m"
elif v == "t2m":
    v_other = "tp"

if flag_another == 1:
    train_other_tmp = obs_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end))[v_other].values.transpose(1,0,2,3)
    train_other_tmp = np.mean(train_other_tmp, axis=1)
    train_other = train_other_tmp[:-num_indent, :, :]

    test_other_tmp = obs_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v_other].values.transpose(1,0,2,3)
    test_other_tmp = np.mean(test_other_tmp, axis=1)
    test_other = np.concatenate([train_other[-num_indent:, :, :], test_other_tmp[:-num_indent, :, :]])

if flag_another_fct == 1:

    train_other_fct = hind_2000_2019_w34.sel(forecast_time=slice(time_train_start,time_train_end))[v_other].values.squeeze().transpose(1,0,2,3)
    train_other_fct = np.nan_to_num(train_other_fct)
    train_other_fct = np.mean(train_other_fct, axis=1)
    train_other_fct = train_other_fct[num_indent:, :, :]

    test_other_fct = hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))[v_other].values.squeeze().transpose(1,0,2,3)
    test_other_fct = np.nan_to_num(test_other_fct)
    test_other_fct = np.mean(test_other_fct, axis=1)

# ---------------------------------------------------------------
if (flag_w56 == 1) and (flag_use_w34 == 1):
    w34_fct_train = hind_2000_2019.isel(lead_time=[0]).sel(forecast_time=slice(time_train_start,time_train_end))[v].values.squeeze().transpose(1,0,2,3)
    w34_fct_test = hind_2000_2019.isel(lead_time=[0]).sel(forecast_time=slice(time_valid_start, time_valid_end))[v].values.squeeze().transpose(1,0,2,3)
    w34_fct_train = np.mean(np.nan_to_num(w34_fct_train), axis=1)[num_indent:, :, :]
    w34_fct_test = np.mean(np.nan_to_num(w34_fct_test), axis=1)

# ---------------------------------------------------------------
### ecmwf parameters
def cal_anom(parameter, flag_w56):
    dates = xr.cftime_range(start='20000102',freq='7D', periods=53)
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        ecmwf_all = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.nc").isel(lead_time=[int(flag_w56)])
        ecmwf_param = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        train_shape = ecmwf_param.to_array().squeeze().shape
        avgs = np.zeros((len(dates),train_shape[-2], train_shape[-1]))
    elif parameter in ["gh", "u", "v", "q"]:
        ecmwf_all = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)])
        ecmwf_param = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
        train_shape = ecmwf_param.to_array().squeeze().shape
        avgs = np.zeros((len(dates), train_shape[-3], train_shape[-2], train_shape[-1]))
    
    if parameter in ["sm100", "sm20", "st20"]:
        ecmwf_param = ecmwf_param.mean("depth_below_and_layer")
        ecmwf_test = ecmwf_test.mean("depth_below_and_layer")
    if parameter in ["tcc", "tcw"]:
        ecmwf_param = ecmwf_param.mean("entireAtmosphere")
        ecmwf_test = ecmwf_test.mean("entireAtmosphere")
    if parameter == "msl":
        ecmwf_param = ecmwf_param.mean("meanSea")
        ecmwf_test = ecmwf_test.mean("meanSea")
    for i,f in enumerate(dates):
        date_range = pd.date_range(start=f.strftime('%Y-%m-%d'),
                    end=f'{time_train_end}-' +  f'{f.month:02d}' + '-' + str(f.day), freq=pd.DateOffset(years=1)).to_list()
        avgs[i] = ecmwf_param.sel(forecast_time=date_range).mean('realization').mean('forecast_time').to_array().squeeze()
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        avgs_ext = np.tile(avgs.transpose([1,2,0]), int(int(time_train_end)-int(time_train_start)+1)).transpose([2,0,1])
        avgs_val = np.tile(avgs.transpose([1,2,0]), int(int(time_valid_end)-int(time_valid_start)+1)).transpose([2,0,1])
    elif parameter in ["gh", "u", "v", "q"]:
        avgs_ext = np.tile(avgs.transpose([1,2,3,0]), int(int(time_train_end)-int(time_train_start)+1)).transpose([3,0,1,2])
        avgs_val = np.tile(avgs.transpose([1,2,3,0]), int(int(time_valid_end)-int(time_valid_start)+1)).transpose([3,0,1,2])
    anom_train = ecmwf_param.mean('realization') - avgs_ext
    anom_test = ecmwf_test.mean('realization') - avgs_val
    anom_train, anom_test = anom_train.to_array().squeeze(), anom_test.to_array().squeeze()
    anom_train = anom_train[num_indent:]
    return anom_train.values, anom_test.values

def load_ecmwf(parameter, flag_w56):
    if parameter in ["rsn", "sst", "sm100", "ci", "msl", "sm20", "st20", "tcc", "tcw"]:
        ecmwf_all = xr.open_dataset(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.nc").isel(lead_time=[int(flag_w56)]).mean('realization')
        ecmwf_train = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
    elif parameter in ["gh", "u", "v", "q"]:
        ecmwf_all = xr.open_zarr(f"{ecmwf_path}/{parameter}_ecmwf_hindcast_2000-2019_biweekly_agg.zarr").isel(lead_time=[int(flag_w56)]).mean('realization')
        ecmwf_train = ecmwf_all.sel(forecast_time=slice(time_train_start,time_train_end))
        ecmwf_test = ecmwf_all.sel(forecast_time=slice(time_valid_start, time_valid_end))
    if parameter in ["sm100", "sm20", "st20"]:
        ecmwf_train = ecmwf_train.mean("depth_below_and_layer")
        ecmwf_test = ecmwf_test.mean("depth_below_and_layer")
    if parameter == "msl":
        ecmwf_param = ecmwf_param.mean("meanSea")
        ecmwf_test = ecmwf_test.mean("meanSea")
    if parameter in ["tcc", "tcw"]:
        ecmwf_param = ecmwf_param.mean("entireAtmosphere")
        ecmwf_test = ecmwf_test.mean("entireAtmosphere")
    ecmwf_train = ecmwf_train.to_array().squeeze()
    ecmwf_test = ecmwf_test.to_array().squeeze()
    train_set = ecmwf_train[num_indent:]
    test_set = ecmwf_test
    return train_set.values, test_set.values

if (flag_use_ecmwf == 1) and (flag_ecmwf_anom != 1):
    if ecmwf_rsn == 1:
        ecmwf_rsn_train, ecmwf_rsn_test = load_ecmwf("rsn", flag_w56)
    if ecmwf_sm100 == 1:
        ecmwf_sm100_train, ecmwf_sm100_test = load_ecmwf("sm100", flag_w56)
    if ecmwf_gh == 1:
        ecmwf_gh_train, ecmwf_gh_test = load_ecmwf("gh", flag_w56)
    if ecmwf_ci == 1:
        ecmwf_ci_train, ecmwf_ci_test = load_ecmwf("ci", flag_w56)
    if ecmwf_u == 1:
        ecmwf_u_train, ecmwf_u_test = load_ecmwf("u", flag_w56)
    if ecmwf_v == 1:
        ecmwf_v_train, ecmwf_v_test = load_ecmwf("v", flag_w56)
    if ecmwf_msl == 1:
        ecmwf_msl_train, ecmwf_msl_test = load_ecmwf("msl", flag_w56)
    if ecmwf_st20 == 1:
        ecmwf_st20_train, ecmwf_st20_test = load_ecmwf("st20", flag_w56)
    if ecmwf_tcc == 1:
        ecmwf_tcc_train, ecmwf_tcc_test = load_ecmwf("tcc", flag_w56)
    if ecmwf_tcw == 1:
        ecmwf_tcw_train, ecmwf_tcw_test = load_ecmwf("tcw", flag_w56)
    if ecmwf_sm20 == 1:
        ecmwf_sm20_train, ecmwf_sm20_test = load_ecmwf("sm20", flag_w56)
    if ecmwf_q == 1:
        ecmwf_q_train, ecmwf_q_test = load_ecmwf("q", flag_w56)
elif (flag_use_ecmwf == 1) and (flag_ecmwf_anom == 1):
    if ecmwf_rsn == 1:
        ecmwf_rsn_train, ecmwf_rsn_test = cal_anom("rsn", flag_w56)
    if ecmwf_sm100 == 1:
        ecmwf_sm100_train, ecmwf_sm100_test = cal_anom("sm100", flag_w56)
    if ecmwf_gh == 1:
        ecmwf_gh_train, ecmwf_gh_test = cal_anom("gh", flag_w56)
    if ecmwf_ci == 1:
        ecmwf_ci_train, ecmwf_ci_test = cal_anom("ci", flag_w56)
    if ecmwf_u == 1:
        ecmwf_u_train, ecmwf_u_test = cal_anom("u", flag_w56)
    if ecmwf_v == 1:
        ecmwf_v_train, ecmwf_v_test = cal_anom("v", flag_w56)
    if ecmwf_msl == 1:
        ecmwf_msl_train, ecmwf_msl_test = cal_anom("msl", flag_w56)
    if ecmwf_st20 == 1:
        ecmwf_st20_train, ecmwf_st20_test = cal_anom("st20", flag_w56)
    if ecmwf_tcc == 1:
        ecmwf_tcc_train, ecmwf_tcc_test = cal_anom("tcc", flag_w56)
    if ecmwf_tcw == 1:
        ecmwf_tcw_train, ecmwf_tcw_test = cal_anom("tcw", flag_w56)
    if ecmwf_sm20 == 1:
        ecmwf_sm20_train, ecmwf_sm20_test = cal_anom("sm20", flag_w56)
    if ecmwf_q == 1:
        ecmwf_q_train, ecmwf_q_test = cal_anom("q", flag_w56)        

region = pd.read_csv(f'{root_path}/data/Regions.txt', sep=',')

d_region = {}
for i in range(region.shape[0]):
    d_region[region.iloc[i]["Region"]] = d_region.get(region.iloc[i]["Region"], [])+[list(region.iloc[i, 1:])]

###################################################################

if flag_preprocess == 1:

    print("!!!!!!!! Start Preprocessing !!!!!!!!")

    ### loop over regions
    for name, loc_region in d_region.items():

        # if os.path.exists(root_dir+"/model/region_model_"+name+".sav"):
        #     print(name, "exit!!")
        #     continue

        print("!!!!!!!! Start >> "+name)

        X_train_all = []
        y_train_all = []

        u1, d1, l1, r1 = loc_region[0]
        u2, d2, l2, r2 = u1, d1, l1, r1
        if len(loc_region) > 1:
            u2, d2, l2, r2 = loc_region[1]

        ### loop over loc
        for idx in range(len(mask_flt[0])):
        # for idx in range(10):

            if idx % 1000 == 0:
                print("Training", idx)

            i = mask_flt[0][idx]
            j = mask_flt[1][idx]  

            ### check whether loc in region
            if ((i>=u1) and (i<d1) and (j>=l1) and (j<r1)) or ((i>=u2) and (i<d2) and (j>=l2) and (j<r2)):

                window = list(np.arange(num_indent, num_indent+9))
                lower = train_terc[num_indent:,0,i,j]
                upper = train_terc[num_indent:,1,i,j]

                ### minmax train sequence
                train_y_avg_rm_scale_mm = ((train_y_avg_rm_scale[:,i,j]-lower)/((upper-lower)+smallv))
                fct_train_mean_part = (fct_train_mean[:,i, j]-lower)/((upper-lower)+smallv)
                fct_train_std_part = (fct_train_std[:,i, j]-lower)/((upper-lower)+smallv)

                ### lag value
                pst_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], len(window)])
                for lgth_idx in range(len(window)):
                    rll = np.roll(train_y_avg_rm_scale_mm, window[lgth_idx])
                    pst_part[:, lgth_idx] = rll

                ### past N year value
                if flag_past_year == 1:
                    ## one year
                    if one_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], 1])
                        rll = np.roll(train_y_avg_rm_scale_mm, 53)
                        year_part[:, 0] = rll

                    ## N years
                    if N_year_past == 1:
                        year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], N_years])
                        for N_year in range(N_years):
                            rll_N = np.roll(train_y_avg_rm_scale_mm, 53*(N_year+1))
                            year_part[:, N_year] = rll_N
                
                #-----------------------------------------------------------------------------------------------    
                ### Training data
                if flag_past_day == 1:
                    exog_set_train = np.concatenate([pst_part, 
                                                     fct_train_mean_part.reshape(-1,1), 
                                                     fct_train_std_part.reshape(-1,1)], axis=1)
                else:
                    exog_set_train = np.concatenate([fct_train_mean_part.reshape(-1,1), 
                                                      fct_train_std_part.reshape(-1,1)], axis=1)
                if flag_past_year == 1:
                    exog_set_train = np.concatenate([exog_set_train, year_part], axis=1)
                
                if flag_fct_bias == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                     bias_mean_train[:, i, j].reshape(-1,1), 
                                                     bias_std_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_use_w34 == 1:
                    if w34_pred == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                        w34_pred_train[:,:,i,j]], axis=1)                
                    if w34_fct == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                            w34_fct_train[:,i,j].reshape(-1,1)], axis=1) 
                #-----------------------------------------------------------------------------------------------    
                if flag_biw[0] == 1:
                    exog_set_train = np.concatenate([exog_set_train, avg_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[1] == 1:
                    exog_set_train = np.concatenate([exog_set_train, std_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[2] == 1:
                    exog_set_train = np.concatenate([exog_set_train, skw_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[3] == 1:
                    exog_set_train = np.concatenate([exog_set_train, krt_train[:, i, j].reshape(-1,1)], axis=1)
                if flag_biw[4] == 1:
                    exog_set_train = np.concatenate([exog_set_train, med_train[:, i, j].reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------    
                if flag_nino == 1:                    
                    exog_set_train = np.concatenate([exog_set_train, 
                                                     nin12_train.reshape(-1,1), nin34_train.reshape(-1,1)], axis=1)
                if flag_solar == 1:
                    exog_set_train = np.concatenate([exog_set_train, solar_train.reshape(-1,1)], axis=1)
                if flag_nao == 1:
                    exog_set_train = np.concatenate([exog_set_train, nao_train.reshape(-1,1)], axis=1)
                if flag_pdo == 1:
                    exog_set_train = np.concatenate([exog_set_train, pdo_train.reshape(-1,1)], axis=1)
                if flag_gl == 1:
                    exog_set_train = np.concatenate([exog_set_train, gl_train.reshape(-1,1)], axis=1)

                if flag_qbo == 1:
                    if flag_qbo_col == -1:
                        exog_set_train = np.concatenate([exog_set_train, qbo_train], axis=1)
                    else:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         qbo_train.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)

                #-----------------------------------------------------------------------------------------------     
                if flag_another == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                     train_other[:, i, j].reshape(-1,1)], axis=1)
                if flag_another_fct == 1:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                     train_other_fct[:, i, j].reshape(-1,1)], axis=1)    
                #-----------------------------------------------------------------------------------------------    
                if flag_use_ecmwf == 1:
                    if ecmwf_rsn == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_rsn_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm100 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_sm100_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_gh == 1:
                        if gh_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_gh_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                             ecmwf_gh_train[:,gh_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_ci == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_ci_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_u == 1:
                        if u_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_u_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                             ecmwf_u_train[:,u_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_v == 1:
                        if v_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_v_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                             ecmwf_v_train[:,v_col,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_msl == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_msl_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_st20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_st20_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcc == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_tcc_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_tcw == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_tcw_train[:,i,j].reshape(-1,1)], axis=1)
                    if ecmwf_sm20 == 1:
                        exog_set_train = np.concatenate([exog_set_train, 
                                                         ecmwf_sm20_train[:,i,j].reshape(-1,1)], axis=1)                
                    if ecmwf_q == 1:
                        if q_col == -1:
                            exog_set_train = np.concatenate([exog_set_train, ecmwf_q_train[:,:,i,j]], axis=1)
                        else:
                            exog_set_train = np.concatenate([exog_set_train, 
                                                             ecmwf_q_train[:,q_col,i,j].reshape(-1,1)], axis=1)                
                #-----------------------------------------------------------------------------------------------    

                train_loc_p = train_seq_p.to_array().values.squeeze()[:, :, i, j]
                train_loc_quant = np.argmax(train_loc_p, axis=0)

                ### cut X_train due to N-past year values
                if (flag_past_year == 1) and (flag_use_w34 != 1):
                    if one_year_past == 1:
                        X_train = exog_set_train[(53-num_indent):]
                        y_train = train_loc_quant[53:]-1
                    if N_year_past == 1:
                        X_train = exog_set_train[((53-num_indent)+53*(N_years-1)):]
                        y_train = train_loc_quant[53*N_years:]-1
                elif (flag_use_w34 == 1) and (w34_pred == 1):
                    X_train = exog_set_train[-(w34_train.shape[0]):]
                    y_train = train_loc_quant[-(w34_train.shape[0]):]-1
                else:
                    X_train = exog_set_train[(53-num_indent):]
                    y_train = train_loc_quant[53:]-1

                X_train_all.extend(X_train)
                y_train_all.extend(y_train)

        X_train_all = np.array(X_train_all)
        y_train_all = np.array(y_train_all)

        print(X_train_all.shape)
        print(y_train_all.shape)
        X_train_all = np.nan_to_num(X_train_all)
        y_train_all = np.nan_to_num(y_train_all)

        ### classification
        if y_train_all.shape[0] != 0:
            print("!!!!!!!! Start classification >> "+name)

            if CLS == "RF":
                clf = RandomForestClassifier(n_estimators=nest, ccp_alpha=ccalpha, max_depth=max_depth, criterion=criterion)
            else:
                print("Not implemented!!")

            clf.fit(X_train_all, y_train_all)

            ### save the model to disk
            save_model_path = root_dir+"/model"
            if not os.path.exists(save_model_path):
                os.makedirs(save_model_path)
            filename = save_model_path+'/region_model_'+name+'.sav'
            pickle.dump(clf, open(filename, 'wb'))
                
            print("!!!!!!!! End classification >> "+name)

    print(">>>>>>>>>>>>>> Finish Overall <<<<<<<<<<<<<")

# ------------------------------------------------------------

if flag_pred == 1:

    print("!!!!!!!! Start Predicting !!!!!!!!")

    for idx in range(len(mask_flt[0])):

        if idx % 1000 == 0:
            print("Testing:", idx)

        i = mask_flt[0][idx]
        j = mask_flt[1][idx]  

        ### find corresponding model
        for name, loc_region in d_region.items():

            u1, d1, l1, r1 = loc_region[0]
            u2, d2, l2, r2 = u1, d1, l1, r1
            if len(loc_region) > 1:
                u2, d2, l2, r2 = loc_region[1]

            if ((i>=u1) and (i<d1) and (j>=l1) and (j<r1)) or ((i>=u2) and (i<d2) and (j>=l2) and (j<r2)):
                filename = root_dir+'/model/region_model_'+name+'.sav'
                clf = pickle.load(open(filename, 'rb'))
                break

        # ### Form training set
        window = list(np.arange(num_indent, num_indent+9))
        lower = train_terc[num_indent:,0,i,j]
        upper = train_terc[num_indent:,1,i,j]

        train_y_avg_rm_scale_mm = ((train_y_avg_rm_scale[:,i,j]-lower)/((upper-lower)+smallv))
        fct_train_mean_part = (fct_train_mean[:,i, j]-lower)/((upper-lower)+smallv)
        fct_train_std_part = (fct_train_std[:,i, j]-lower)/((upper-lower)+smallv)

        ### lag value
        pst_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], len(window)])
        for lgth_idx in range(len(window)):
            rll = np.roll(train_y_avg_rm_scale_mm, window[lgth_idx])
            pst_part[:, lgth_idx] = rll

        ### past N year value
        if flag_past_year == 1:
            ## one year
            if one_year_past == 1:
                year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], 1])
                rll = np.roll(train_y_avg_rm_scale_mm, 53)
                year_part[:, 0] = rll

            ## two year
            if N_year_past == 1:
                year_part = np.zeros([train_y_avg_rm_scale_mm.shape[0], N_years])
                for N_year in range(N_years):
                    rll_N = np.roll(train_y_avg_rm_scale_mm, 53*(N_year+1))
                    year_part[:, N_year] = rll_N
        
        #-----------------------------------------------------------------------------------------------    
        ### Training data
        if flag_past_day == 1:
            exog_set_train = np.concatenate([pst_part, 
                                            fct_train_mean_part.reshape(-1,1), 
                                            fct_train_std_part.reshape(-1,1)], axis=1)
        else:
            exog_set_train = np.concatenate([fct_train_mean_part.reshape(-1,1), 
                                                fct_train_std_part.reshape(-1,1)], axis=1)
        if flag_past_year == 1:
            exog_set_train = np.concatenate([exog_set_train, year_part], axis=1)
        
        if flag_fct_bias == 1:
            exog_set_train = np.concatenate([exog_set_train, 
                                                bias_mean_train[:, i, j].reshape(-1,1), 
                                                bias_std_train[:, i, j].reshape(-1,1)], axis=1)

        #-----------------------------------------------------------------------------------------------    
        if flag_use_w34 == 1:
            if w34_pred == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                w34_pred_train[:,:,i,j]], axis=1)                
            if w34_fct == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                    w34_fct_train[:,i,j].reshape(-1,1)], axis=1) 
        #-----------------------------------------------------------------------------------------------    
        if flag_biw[0] == 1:
            exog_set_train = np.concatenate([exog_set_train, avg_train[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[1] == 1:
            exog_set_train = np.concatenate([exog_set_train, std_train[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[2] == 1:
            exog_set_train = np.concatenate([exog_set_train, skw_train[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[3] == 1:
            exog_set_train = np.concatenate([exog_set_train, krt_train[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[4] == 1:
            exog_set_train = np.concatenate([exog_set_train, med_train[:, i, j].reshape(-1,1)], axis=1)

        #-----------------------------------------------------------------------------------------------    
        if flag_nino == 1:                    
            exog_set_train = np.concatenate([exog_set_train, 
                                                nin12_train.reshape(-1,1), nin34_train.reshape(-1,1)], axis=1)
        if flag_solar == 1:
            exog_set_train = np.concatenate([exog_set_train, solar_train.reshape(-1,1)], axis=1)
        if flag_nao == 1:
            exog_set_train = np.concatenate([exog_set_train, nao_train.reshape(-1,1)], axis=1)
        if flag_pdo == 1:
            exog_set_train = np.concatenate([exog_set_train, pdo_train.reshape(-1,1)], axis=1)
        if flag_gl == 1:
            exog_set_train = np.concatenate([exog_set_train, gl_train.reshape(-1,1)], axis=1)

        if flag_qbo == 1:
            if flag_qbo_col == -1:
                exog_set_train = np.concatenate([exog_set_train, qbo_train], axis=1)
            else:
                exog_set_train = np.concatenate([exog_set_train, 
                                                 qbo_train.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)
        if flag_another == 1:
            exog_set_train = np.concatenate([exog_set_train, 
                                                train_other[:, i, j].reshape(-1,1)], axis=1)
        if flag_another_fct == 1:
            exog_set_train = np.concatenate([exog_set_train, 
                                                train_other_fct[:, i, j].reshape(-1,1)], axis=1)    

        #-----------------------------------------------------------------------------------------------    
        if flag_use_ecmwf == 1:
            if ecmwf_rsn == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                 ecmwf_rsn_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_sm100 == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                 ecmwf_sm100_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_gh == 1:
                if gh_col == -1:
                    exog_set_train = np.concatenate([exog_set_train, ecmwf_gh_train[:,:,i,j]], axis=1)
                else:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_gh_train[:,gh_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_ci == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                    ecmwf_ci_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_u == 1:
                if u_col == -1:
                    exog_set_train = np.concatenate([exog_set_train, ecmwf_u_train[:,:,i,j]], axis=1)
                else:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_u_train[:,u_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_v == 1:
                if v_col == -1:
                    exog_set_train = np.concatenate([exog_set_train, ecmwf_v_train[:,:,i,j]], axis=1)
                else:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_v_train[:,v_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_msl == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                ecmwf_msl_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_st20 == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                ecmwf_st20_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_tcc == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                ecmwf_tcc_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_tcw == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                ecmwf_tcw_train[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_sm20 == 1:
                exog_set_train = np.concatenate([exog_set_train, 
                                                ecmwf_sm20_train[:,i,j].reshape(-1,1)], axis=1)      
            if ecmwf_q == 1:
                if q_col == -1:
                    exog_set_train = np.concatenate([exog_set_train, ecmwf_q_train[:,:,i,j]], axis=1)
                else:
                    exog_set_train = np.concatenate([exog_set_train, 
                                                        ecmwf_q_train[:,q_col,i,j].reshape(-1,1)], axis=1)        
        #-----------------------------------------------------------------------------------------------   
        #  
        ### cut X_train due to N-past year values
        if (flag_past_year == 1) and (flag_use_w34 != 1):
            if one_year_past == 1:
                X_train = exog_set_train[(53-num_indent):]
            if N_year_past == 1:
                X_train = exog_set_train[((53-num_indent)+53*(N_years-1)):]
        elif (flag_use_w34 == 1) and (w34_pred == 1):
            X_train = exog_set_train[-(w34_train.shape[0]):]
        else:
            X_train = exog_set_train[(53-num_indent):]

        X_train = np.nan_to_num(X_train)

        y_train_prob = clf.predict_proba(X_train)
        save_train_prob_path = root_dir+"/y_train_prob_"+str(n)
        if not os.path.exists(save_train_prob_path):
            os.makedirs(save_train_prob_path)
        np.save(save_train_prob_path+"/y_train_prob_"+str(i)+"by"+str(j)+"_"+str(n)+".npy", y_train_prob)


        # ########################################
        ### testing

        test_lower = test_terc[:,0,i,j]
        test_upper = test_terc[:,1,i,j]
        test_y_avg_rm_scale_mm = ((test_y_avg_rm_scale[:,i,j]-test_lower)/((test_upper-test_lower)+smallv))
        fct_test_mean_part = (fct_test_mean[:,i, j]-test_lower)/((test_upper-test_lower)+smallv)
        fct_test_std_part = (fct_test_std[:,i, j]-test_lower)/((test_upper-test_lower)+smallv)
        
        ### past days value
        roll_all = np.concatenate([train_y_avg_rm_scale_mm[-len(window):], test_y_avg_rm_scale_mm], axis=0)
        pst_test_part = np.zeros([roll_all.shape[0], len(window)])
        for lgth_idx in range(len(window)):
            rll = np.roll(roll_all, window[lgth_idx])
            pst_test_part[:, lgth_idx] = rll
        pst_test_part = pst_test_part[-len(test_y_avg_rm_scale_mm):]

        ### past years (same time) value
        if flag_past_year == 1:
            if one_year_past == 1:
                year_test_part = train_y_avg_rm_scale_mm[-53:].reshape(-1,1)
            if N_year_past == 1:
                year_test_part = np.zeros([53, N_years])
                for N_year in range(N_years):
                    if N_year == 0:
                        year_test_part[:,0] = train_y_avg_rm_scale_mm[-53:]
                    else:
                        year_test_part[:,N_year] = train_y_avg_rm_scale_mm[-53*(N_year+1):-53*N_year]
        
        #-----------------------------------------------------------------------------------------------    
        ### Testing data
        if flag_past_day == 1:
            exog_set_test = np.concatenate([pst_test_part, 
                                            fct_test_mean_part.reshape(-1,1), 
                                            fct_test_std_part.reshape(-1,1)], axis=1)
        else:
            exog_set_test = np.concatenate([fct_test_mean_part.reshape(-1,1), 
                                            fct_test_std_part.reshape(-1,1)], axis=1)
        if flag_past_year == 1:
            exog_set_test = np.concatenate([exog_set_test, year_test_part], axis=1)
        
        if flag_fct_bias == 1:
            exog_set_test = np.concatenate([exog_set_test, 
                                            bias_mean_test[:, i, j].reshape(-1,1), 
                                            bias_std_test[:, i, j].reshape(-1,1)], axis=1)

        #-----------------------------------------------------------------------------------------------    
        if flag_use_w34 == 1:
            if w34_pred == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                   w34_pred_test[:,:,i,j]], axis=1)                
            if w34_fct == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                    w34_fct_test[:,i,j].reshape(-1,1)], axis=1) 
        #-----------------------------------------------------------------------------------------------    
        if flag_biw[0] == 1:
            exog_set_test = np.concatenate([exog_set_test, avg_test[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[1] == 1:
            exog_set_test = np.concatenate([exog_set_test, std_test[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[2] == 1:
            exog_set_test = np.concatenate([exog_set_test, skw_test[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[3] == 1:
            exog_set_test = np.concatenate([exog_set_test, krt_test[:, i, j].reshape(-1,1)], axis=1)
        if flag_biw[4] == 1:
            exog_set_test = np.concatenate([exog_set_test, med_test[:, i, j].reshape(-1,1)], axis=1)

        #-----------------------------------------------------------------------------------------------    
        if flag_nino == 1:                    
            exog_set_test = np.concatenate([exog_set_test, 
                                            nin12_test.reshape(-1,1), 
                                            nin34_test.reshape(-1,1)], axis=1)
        if flag_solar == 1:
            exog_set_test = np.concatenate([exog_set_test, solar_test.reshape(-1,1)], axis=1)
        if flag_nao == 1:
            exog_set_test = np.concatenate([exog_set_test, nao_test.reshape(-1,1)], axis=1)
        if flag_pdo == 1:
            exog_set_test = np.concatenate([exog_set_test, pdo_test.reshape(-1,1)], axis=1)
        if flag_gl == 1:
            exog_set_test = np.concatenate([exog_set_test, gl_test.reshape(-1,1)], axis=1)

        if flag_qbo == 1:
            if flag_qbo_col == -1:
                exog_set_test = np.concatenate([exog_set_test, qbo_test], axis=1)
            else:
                exog_set_test = np.concatenate([exog_set_test, 
                                                qbo_test.iloc[:,flag_qbo_col].values.reshape(-1,1)], axis=1)
        #-----------------------------------------------------------------------------------------------    
        if flag_another == 1:
            exog_set_test = np.concatenate([exog_set_test, 
                                            test_other[:, i, j].reshape(-1,1)], axis=1)
        if flag_another_fct == 1:
            exog_set_test = np.concatenate([exog_set_test, 
                                            test_other_fct[:, i, j].reshape(-1,1)], axis=1)    

        #-----------------------------------------------------------------------------------------------    
        if flag_use_ecmwf == 1:
            if ecmwf_rsn == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_rsn_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_sm100 == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_sm100_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_gh == 1:
                if gh_col == -1:
                    exog_set_test = np.concatenate([exog_set_test, ecmwf_gh_test[:,:,i,j]], axis=1)
                else:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    ecmwf_gh_test[:,gh_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_ci == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_ci_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_u == 1:
                if u_col == -1:
                    exog_set_test = np.concatenate([exog_set_test, ecmwf_u_test[:,:,i,j]], axis=1)
                else:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    ecmwf_u_test[:,u_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_v == 1:
                if v_col == -1:
                    exog_set_test = np.concatenate([exog_set_test, ecmwf_v_test[:,:,i,j]], axis=1)
                else:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    ecmwf_v_test[:,v_col,i,j].reshape(-1,1)], axis=1)
            if ecmwf_msl == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_msl_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_st20 == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_st20_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_tcc == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_tcc_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_tcw == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_tcw_test[:,i,j].reshape(-1,1)], axis=1)
            if ecmwf_sm20 == 1:
                exog_set_test = np.concatenate([exog_set_test, 
                                                ecmwf_sm20_test[:,i,j].reshape(-1,1)], axis=1)      
            if ecmwf_q == 1:
                if q_col == -1:
                    exog_set_test = np.concatenate([exog_set_test, ecmwf_q_test[:,:,i,j]], axis=1)
                else:
                    exog_set_test = np.concatenate([exog_set_test, 
                                                    ecmwf_q_test[:,q_col,i,j].reshape(-1,1)], axis=1)
        #-----------------------------------------------------------------------------------------------    
        
        ### form X_test
        X_test = exog_set_test
        X_test = np.nan_to_num(X_test)
        
        y_test_prob = clf.predict_proba(X_test)
        save_seq_prob_path = root_dir+"/y_test_prob_"+str(n)
        if not os.path.exists(save_seq_prob_path):
            os.makedirs(save_seq_prob_path)
        np.save(save_seq_prob_path+"/y_test_prob_"+str(i)+"by"+str(j)+"_"+str(n)+".npy", y_test_prob)


    print("!!!!!!!! End Predicting !!!!!!!!")
    print(root_dir)

# ------------------------------------------------------------
