########################################################################
# Script for combination per location prediction to one file
### both train and test

# To run the script:
#   python3 pred2rpss_t2m.py argv1 argv2
#   @param argv1: "w34" for week 3-4 prediction
#                 "w56" for week 5-6 prediction
#   @param argv2: 1 for train
#                 0 for test
######################################################################

from sklearn.preprocessing import OneHotEncoder
import sys, pickle
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import matplotlib.patches as patches
from textwrap import wrap

import xskillscore as xs
from os import listdir
from os.path import isfile, join

from scripts_tp import *

# --------------------------------------------

### predict week 3-4 or week 5-6
wk = sys.argv[1]
flag_use_train = int(sys.argv[2])
score_type=sys.argv[3]

cache_path = "../../data/newdata"
v = 'tp'
regional_score = True

time_valid_start = sys.argv[4]
time_valid_end = str(int(time_valid_start)+1)


# # ---------------------------------------------

root_dir = f"../result_sys_valid_tp_ecmwf_tmp1/cv/cv_{time_valid_start}_{time_valid_end}/"

print("-------------------------------------------")
print("\n>>>>>>>>>", root_dir, "<<<<<<<<<<<\n")

##===========================================================

if wk == "w34":
    flag_w34 = 1                            ### predict w34
    flag_w56 = 0
    criterion = 'gini'                      ### entropy for w34, gini for w56

elif wk == "w56":
    flag_w56 = 1                            ### predict w56
    flag_w34 = 0
    criterion = 'entropy'                      ### entropy for w34, gini for w56
    
# # ---------------------------------------------

flag_prob2all = 1
flag_pred2all = 1-flag_prob2all
flag_pred2oneHot = flag_pred2all
flag_get_RPSS = 1
flag_rpssMap = 1

##===========================================================
### Load data
if flag_w34 == 1:
    ### train output
    try:
        obs_2000_2019_w34 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        obs_2000_2019_w34 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### test output
    try:
        obs_2020_w34 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        obs_2020_w34 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### train input
    try:
        hind_2000_2019_w34 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        hind_2000_2019_w34 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### test input
    try:
        fct_2020_w34 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        fct_2020_w34 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr").isel(lead_time=[0])
elif flag_w56 == 1:
    ### train output
    try:
        obs_2000_2019_w34 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        obs_2000_2019_w34 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### test output
    try:
        obs_2020_w34 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        obs_2020_w34 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### train input
    try:
        hind_2000_2019_w34 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        hind_2000_2019_w34 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### test input
    try:
        fct_2020_w34 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        fct_2020_w34 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr").isel(lead_time=[1])

### tercile
try:
    tercile_edges = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.nc")
except:
    tercile_edges = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.zarr")

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = obs_2020_w34.std(['lead_time','forecast_time']).notnull()[v]

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = obs_2020_w34.std(['lead_time','forecast_time']).notnull()[v]
# --------------------------------------------

train_data = obs_2000_2019_w34.sel(forecast_time=slice('2000', '2019'))[v].values.transpose(1,0,2,3)
train_data = np.mean(train_data, axis=1)

mask_flt = np.where(~np.isnan(train_data[316]))

# --------------------------------------------

if flag_prob2all == 1:

    if flag_use_train == 1:
        if score_type == "train":
            prob_path = root_dir+"/y_train_prob_10"
        elif score_type == "valid":
            prob_path = root_dir+"/y_test_prob_10"
    else:
        prob_path = root_dir+"/y_2020_prob_10"

    # print(prob_path)

    prob_file = [f for f in listdir(prob_path) if isfile(join(prob_path, f))]

    tmp = np.load(prob_path+"/"+prob_file[0])
    # print(tmp.shape)

    pred_prob_arry_all = np.zeros([tmp.shape[0], 3, 121, 240])

    for file in prob_file:
        # print(file)

        i = int(file.split("by")[0].split("_")[-1])
        j = int(file.split("by")[1].split("_")[0])

        pred_prob_arry = np.load(prob_path+"/"+file)
        pred_prob_arry_all[:,:,i,j] = pred_prob_arry

    pred_prob_arry_all = np.transpose(pred_prob_arry_all, (1,0,2,3))

    if flag_use_train == 1:
        if score_type == "train":
            np.savez(root_dir+"/train_prob_AR_10.npz", pred_prob_arry_all)
        elif score_type == "valid":
            np.savez(root_dir+"/valid_prob_AR_10.npz", pred_prob_arry_all)
    else:
        np.savez(root_dir+"/test_2020_prob_AR_10.npz", pred_prob_arry_all)


if flag_pred2all == 1:

    if flag_use_train == 1:
        pred_path = root_dir+"/y_pred_10"
    else:
        pred_path = root_dir+"/pred_seq_10"
    
    pred_file = [f for f in listdir(pred_path) if isfile(join(pred_path, f))]
    tmp = np.load(pred_path+"/"+pred_file[0])
    pred_arry_all = np.zeros([tmp.shape[0], 121, 240])

    for file in pred_file:
        i = int(file.split("by")[0].split("_")[-1])
        j = int(file.split("by")[1].split("_")[0])
        pred_arry = np.load(pred_path+"/"+file)
        pred_arry_all[:,i,j] = pred_arry

    if flag_use_train == 1:
        if score_type == "train":
            np.savez(root_dir+"/train_pred_AR_10.npz", pred_arry_all)
        elif score_type == "valid":
            np.savez(root_dir+"/valid_pred_AR_10.npz", pred_arry_all)
    else:
        np.savez(root_dir+"/test_2020_pred_AR_10.npz", pred_arry_all)


# ------------------------------

if flag_pred2oneHot == 1:

    if flag_use_train == 1:
        if score_type == "train":
            pred_HMM = np.load(root_dir+"/train_pred_AR_10.npz")["arr_0"]
        elif score_type == "valid":
            pred_HMM = np.load(root_dir+"/valid_pred_AR_10.npz")["arr_0"]
    else:
        pred_HMM = np.load(root_dir+"/test_2020_pred_AR_10.npz")["arr_0"]

    oneHot_HMM = np.zeros([3, pred_HMM.shape[0], pred_HMM.shape[1], pred_HMM.shape[2]])
    oneHot_HMM[oneHot_HMM == 0] = 'nan'

    for idx in range(len(mask_flt[0])):

        i = mask_flt[0][idx]
        j = mask_flt[1][idx]  

        ds_seq = pred_HMM[:,i,j]
        
        if -1 not in ds_seq:
            ds_seq = np.append(ds_seq, -1)
        if 0 not in ds_seq:
            ds_seq = np.append(ds_seq, 0)
        if 1 not in ds_seq:
            ds_seq = np.append(ds_seq, 1)

        onehot_encoder = OneHotEncoder(sparse=False)
        integer_encoded = np.array(ds_seq).reshape(len(ds_seq), 1)
        # print(integer_encoded.shape)

        onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
        onehot_encoded = onehot_encoded[:pred_HMM.shape[0], :]
        onehot_encoded = np.transpose(onehot_encoded)

        oneHot_HMM[:,:,i,j] = onehot_encoded

    if flag_use_train == 1:
        if score_type == "train":
            np.savez(root_dir+"/train_oneHot_HMM.npz", oneHot_HMM)
        elif score_type == "valid":
            np.savez(root_dir+"/test_oneHot_HMM.npz", oneHot_HMM)
    else:
        np.savez(root_dir+"/test_2020_oneHot_HMM.npz", oneHot_HMM)


# ------------------------------

def func_get_RPSS(oneHot_HMM, ds_cmpr, time):
    oneHot_HMM_test = np.expand_dims(oneHot_HMM, axis=2)
    ds_cmpr = ds_cmpr.sel(forecast_time=time)
    oneHot_HMM_da = xr.DataArray(oneHot_HMM_test,
                            dims=['category', 'forecast_time', 'lead_time', 'latitude', 'longitude'],
                            coords={'category': ['below normal', 'near normal', 'above normal'],
                                    'forecast_time': ds_cmpr.forecast_time,
                                    'lead_time': ds_cmpr.lead_time,
                                    'latitude': ds_cmpr.latitude,
                                    'longitude': ds_cmpr.longitude},
                                    )
    oneHot_HMM_add = add_valid_time_from_forecast_reference_time_and_lead_time(oneHot_HMM_da)
    oneHot_HMM_p = oneHot_HMM_add.to_dataset(name=v)
    return oneHot_HMM_p

# ------------------------------

if flag_get_RPSS == 1:

    if flag_pred2oneHot == 1:
        if flag_use_train == 1:
            if score_type == "train":
                oneHot_HMM = np.load(root_dir+"/train_oneHot_HMM.npz")["arr_0"]
            elif score_type == "valid":
                oneHot_HMM = np.load(root_dir+"/valid_oneHot_HMM.npz")["arr_0"]
        else:
            oneHot_HMM = np.load(root_dir+"/test_2020_oneHot_HMM.npz")["arr_0"]

    if flag_prob2all == 1:
        if flag_use_train == 1:
            if score_type == "train":
                oneHot_HMM = np.load(root_dir+"/train_prob_AR_10.npz")["arr_0"]
            elif score_type == "valid":
                oneHot_HMM = np.load(root_dir+"/valid_prob_AR_10.npz")["arr_0"]
        else:
            oneHot_HMM = np.load(root_dir+"/test_2020_prob_AR_10.npz")["arr_0"]

    if flag_use_train == 1:
        # time = slice('2000', '2018')
        yr_len = oneHot_HMM.shape[1]//53
        if score_type == "train":
            strt = (int(time_valid_start)-1)-yr_len+1
            time = slice(str(strt), str(int(time_valid_start)-1))
            ds_cmpr = hind_2000_2019_w34.sel(forecast_time=slice(str(strt), str(int(time_valid_start)-1)))
        elif score_type == "valid":
            time = slice(time_valid_start, time_valid_end)
            ds_cmpr = hind_2000_2019_w34.sel(forecast_time=slice(time_valid_start, time_valid_end))
    else:
        time='2020'
        ds_cmpr=fct_2020_w34

    oneHot_HMM_p = func_get_RPSS(oneHot_HMM, ds_cmpr, time)
    score_avg, score = skill_by_year_new(oneHot_HMM_p, cache_path = cache_path, name=v)
    score_avg_all, score_all = skill_by_year_new(oneHot_HMM_p, cache_path = cache_path, name=v, score_all = True)

    # print(score)
    print("(None, -60): ", score_avg.round(6))

# ------------------------------

if flag_rpssMap == 1:

    preds = oneHot_HMM_p
    xr.set_options(keep_attrs=True)
    if flag_use_train == 1:
        try:
            obs_p = xr.open_dataset(f'{cache_path}/hindcast-like-observations_2000-2019_biweekly_terciled.nc').sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)
        except:
            obs_p = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_terciled.zarr").sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)
    else:
        obs_p = xr.open_dataset(f'{cache_path}/forecast-like-observations_2020_biweekly_terciled.nc').sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)

    obs_p = obs_p[v].to_dataset(name=v)

    # ML probabilities
    fct_p = preds

    # climatology
    clim_p = xr.DataArray([1/3, 1/3, 1/3], dims='category', coords={'category':['below normal', 'near normal', 'above normal']}).to_dataset(name=v)

    assert_predictions_2020(obs_p)
    assert_predictions_2020(fct_p)

    ## RPSS
    # rps_ML
    rps_ML = xs.rps(obs_p, fct_p, category_edges=None, dim=[], input_distributions='p').compute()
    # rps_clim
    # rps_clim = xs.rps(obs_p, clim_p, category_edges=None, dim='forecast_time', input_distributions='p').compute()
    rps_clim = xs.rps(obs_p, clim_p, category_edges=None, dim=[], input_distributions='p').compute()
    
    expect = obs_p.sum('category')
    expect = expect.where(expect > 0.98).where(expect < 1.02)  # should be True if not all NaN
    rps_ML = rps_ML.where(expect, other=2)  # assign RPS=2 where value was expected but NaN found

    ### rpss
    rpss = 1 - (rps_ML.groupby('forecast_time.year').mean() / rps_clim.groupby('forecast_time.year').mean())
    # dirty fix
    rpss = rpss.clip(-10, 1)
    
    rrr = rpss.to_array().values.squeeze()

    if flag_use_train == 1:
        rrr = rrr[-1, :, :]
        loc_rpss_sml = np.where(rrr < 0)
        np.save(root_dir+"/train_small_loc.npy", loc_rpss_sml)
    else:
        loc_rpss_sml = np.where(rrr < 0)

    if (regional_score == True):
        weights = np.cos(np.deg2rad(np.abs(rpss.latitude)))
        region = pd.read_csv('../data/Regions.txt', sep=',')
        d_region = {}
        for i in range(region.shape[0]):
            d_region[region.iloc[i]["Region"]] = d_region.get(region.iloc[i]["Region"], [])+[list(region.iloc[i, 1:])]
        d_score = d_region.copy()

        plt.figure(figsize=[10,5])
        plt.imshow(rrr)
        plt.colorbar()
        plt.tight_layout()

        score_region = {}
        df_score_region = pd.DataFrame()
        for name, loc_region in d_region.items():
            u1, d1, l1, r1 = loc_region[0]
            # d1, r1 = d1-1, r1-1
            l1_m, r1_m = l1*1.5, r1*1.5
            u1_m, d1_m = u1*(-1.5)+90, d1*(-1.5)+90
            
            part = rpss.sel(latitude=slice(u1_m, d1_m), longitude=slice(l1_m, r1_m))# .weighted(weights).mean('latitude').mean('longitude')

            rect = patches.Rectangle((l1, u1), r1-l1, d1-u1, linewidth=1, edgecolor='r', facecolor='none')
            plt.gca().add_patch(rect)

            if len(loc_region) > 1:
                u2, d2, l2, r2 = loc_region[1]
                # d2, r2 = d2-1, r2-1
                l2_m, r2_m = l2*1.5, r2*1.5
                u2_m, d2_m = u2*(-1.5)+90, d2*(-1.5)+90

                part2 = rpss.sel(latitude=slice(u2_m, d2_m), longitude=slice(l2_m, r2_m))#.weighted(weights).mean('latitude').mean('longitude')
                part = xr.concat([part, part2], dim="longitude")
        
                rect = patches.Rectangle((l2, u2), r2-l2, d2-u2, linewidth=1, edgecolor='r', facecolor='none')
                plt.gca().add_patch(rect)

            scores  = part.weighted(weights).mean('latitude').mean('longitude')
            scores = scores.to_array().mean(['lead_time', 'variable'])

            t = (name+"\n"+str(round(scores.values[0], 6)))
            plt.text(l1+1, d1-2, t, fontsize=7.8,  wrap=True)

            d_score[name].append(round(scores.values[0], 6))
            score_region[name] = round(scores.values[0], 6)
            df_score_region = pd.concat([df_score_region, pd.DataFrame([name, scores.values[0]]).T.set_index(0)])
        
        # if score_type == "test":
        #     df_score_region.columns = ["test_score"]
        #     val_score_df = pd.read_csv(f"../result_sys_valid_tp_ecmwf_tmp1/level_var/{var_name}_score_set.csv", header=None, index_col=0)
        #     val_score_df.columns = ["unit_level", "unit_fldr", "val_score"]

        #     score_merge = pd.merge(val_score_df, df_score_region, left_index=True, right_index=True)
        #     score_merge.to_csv(root_dir + f"/{var_name}_scores.csv")

        # f = open(root_dir+"/score_region.pkl", "wb")
        # pickle.dump(d_score,f)
        plt.title("Test RPSS on {}, RPSS={:.5f}".format(score_type, score.values[0][0]))
        feat_set = root_dir.split("/")[-1].split("w560_")[-1]
        plt.text(-8, 132, feat_set, fontsize=8.2, wrap = True)
        plt.savefig(f"{root_dir}/RPSS_{score_type}_region.png")
        plt.close()

    # with open(root_dir+"/rpss_small_loc.txt", "w") as file:
    #     file.write(str(loc_rpss_sml))

    plt.figure(figsize=[10,5])
    plt.imshow(rrr)
    plt.tight_layout()
    plt.colorbar()
    plt.title("Test RPSS on {}, RPSS={:.5f}".format(score_type, score.values[0][0]))
    if flag_prob2all == 1:
        plt.savefig(root_dir+f"/RPSS_{score_type}_prob_clr.png")
    elif flag_pred2all == 1:
        plt.savefig(root_dir+f"/RPSS_{score_type}_pred_clr.png")

    # for idx in range(len(loc_rpss_sml[0])):
    #     plt.scatter(x=loc_rpss_sml[1][idx], y=loc_rpss_sml[0][idx], c='r', s=10)

    # if flag_use_train == 1:
    #     plt.title("Train (last yr) RPSS on TS, RPSS={:.3f}".format(score.values[0][0]))
    #     if flag_prob2all == 1:
    #         plt.savefig(root_dir+"/RPSS_train_prob.png")
    #     elif flag_pred2all == 1:
    #         plt.savefig(root_dir+"/RPSS_train_pred.png")
    # else:
    #     plt.title("Test RPSS on TS, RPSS={:.5f}".format(score.values[0][0]))
    #     if flag_prob2all == 1:
    #         plt.savefig(root_dir+"/RPSS_test_prob.png")
    #     elif flag_pred2all == 1:
    #         plt.savefig(root_dir+"/RPSS_test_pred.png")

    plt.close()

# ---------------------------------------------
