########################################################################
# Script for combination per location prediction to one file
### both train and test

# To run the script:
#   python3 pred2rpss_t2m.py argv1 argv2
#   @param argv1: "w34" for week 3-4 prediction
#                 "w56" for week 5-6 prediction
#   @param argv2: 1 for train
#                 0 for test
######################################################################

from sklearn.preprocessing import OneHotEncoder
import sys, pickle, os
import numpy as np
import xarray as xr

import xskillscore as xs
from os import listdir
from os.path import isfile, join
import matplotlib.pyplot as plt

from scripts_tp import *

# --------------------------------------------

### predict week 3-4 or week 5-6
wk = sys.argv[1]
flag_use_train = int(sys.argv[2])

cache_path = "../../data/newdata"
v = 'tp'

# num_unit = int(sys.argv[4])

if wk == "w34":
    flag_w34 = 1                            ### predict w34
    flag_w56 = 0

elif wk == "w56":
    flag_w56 = 1                            ### predict w56
    flag_w34 = 0

time_train_start, time_train_end = '2000', '2017' # train
time_valid_start, time_valid_end = '2018', '2019' # valid
time_test = '2020'                                # test

# # ---------------------------------------------
# root_path = f"../result_sys_valid_tp_ecmwf/unit{num_unit}/"
# root_dir = f"result_tp_ecmwf/RegModel_RF_gini_w341_w560_ecmwf11_rsn0sm1000sst00-1gh1{sys.argv[3]}"
# root_d = "result_sys_valid_tp_ecmwf/unit2/Europe/"
root_d = sys.argv[3]+"/"

print("-------------------------------------------")
print("\n>>>>>>>>>", root_d, "<<<<<<<<<<<\n")
##===========================================================

flag_prob2all = 1
flag_pred2all = 1-flag_prob2all
flag_pred2oneHot = flag_pred2all
flag_get_RPSS = 1
flag_rpssMap = 1

if v == 't2m':   
    from scripts_t2m import *
else:
    from scripts_tp import *

##===========================================================
### Load data
if flag_w34 == 1:
    ### train output
    try:
        obs_2000_2019_w34 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        obs_2000_2019_w34 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### test output
    try:
        obs_2020_w34 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        obs_2020_w34 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### train input
    try:
        hind_2000_2019_w34 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        hind_2000_2019_w34 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[0])
    ### test input
    try:
        fct_2020_w34 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc").isel(lead_time=[0])
    except:
        fct_2020_w34 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr").isel(lead_time=[0])
elif flag_w56 == 1:
    ### train output
    try:
        obs_2000_2019_w34 = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        obs_2000_2019_w34 = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### test output
    try:
        obs_2020_w34 = xr.open_dataset(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        obs_2020_w34 = xr.open_zarr(f"{cache_path}/forecast-like-observations_2020_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### train input
    try:
        hind_2000_2019_w34 = xr.open_dataset(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        hind_2000_2019_w34 = xr.open_zarr(f"{cache_path}/ecmwf_hindcast-input_2000-2019_biweekly_deterministic.zarr").isel(lead_time=[1])
    ### test input
    try:
        fct_2020_w34 = xr.open_dataset(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.nc").isel(lead_time=[1])
    except:
        fct_2020_w34 = xr.open_zarr(f"{cache_path}/ecmwf_forecast-input_2020_biweekly_deterministic.zarr").isel(lead_time=[1])

### tercile
try:
    tercile_edges = xr.open_dataset(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.nc")
except:
    tercile_edges = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_tercile-edges.zarr")

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = obs_2020_w34.std(['lead_time','forecast_time']).notnull()[v]

### Mask
mask = obs_2020_w34.std(['lead_time','forecast_time']).notnull()
mask_v = obs_2020_w34.std(['lead_time','forecast_time']).notnull()[v]
# --------------------------------------------

train_data = obs_2000_2019_w34.sel(forecast_time=slice(time_train_start, time_train_end))[v].values.transpose(1,0,2,3)
train_data = np.mean(train_data, axis=1)

mask_flt = np.where(~np.isnan(train_data[316]))

# --------------------------------------------

def func_get_RPSS(oneHot_HMM, ds_cmpr, time):
    oneHot_HMM_test = np.expand_dims(oneHot_HMM, axis=2)
    ds_cmpr = ds_cmpr.sel(forecast_time=time)
    oneHot_HMM_da = xr.DataArray(oneHot_HMM_test,
                            dims=['category', 'forecast_time', 'lead_time', 'latitude', 'longitude'],
                            coords={'category': ['below normal', 'near normal', 'above normal'],
                                    'forecast_time': ds_cmpr.forecast_time,
                                    'lead_time': ds_cmpr.lead_time,
                                    'latitude': ds_cmpr.latitude,
                                    'longitude': ds_cmpr.longitude},
                                    )
    oneHot_HMM_add = add_valid_time_from_forecast_reference_time_and_lead_time(oneHot_HMM_da)
    oneHot_HMM_p = oneHot_HMM_add.to_dataset(name=v)
    return oneHot_HMM_p

# --------------------------------------------
region = pd.read_csv('../data/Regions.txt', sep=',')
d_region = {}
for i in range(region.shape[0]):
    d_region[region.iloc[i]["Region"]] = d_region.get(region.iloc[i]["Region"], [])+[list(region.iloc[i, 1:])]
d_score = d_region.copy()
name = root_d.split("/")[-2]
dd = pd.DataFrame()

# --------------------------------------------

for filename in os.listdir(root_d):
    root_dir = os.path.join(root_d, filename)

    # ---------------

    if flag_prob2all == 1:

        if flag_use_train == 1:
            prob_path = root_dir+"/y_train_prob_10"
        else:
            prob_path = root_dir+"/y_test_prob_10"

        prob_file = [f for f in listdir(prob_path) if isfile(join(prob_path, f))]

        tmp = np.load(prob_path+"/"+prob_file[0])
        # print(tmp.shape)

        pred_prob_arry_all = np.zeros([tmp.shape[0], 3, 121, 240])

        for file in prob_file:

            i = int(file.split("by")[0].split("_")[-1])
            j = int(file.split("by")[1].split("_")[0])

            pred_prob_arry = np.load(prob_path+"/"+file)
            pred_prob_arry_all[:,:,i,j] = pred_prob_arry

        pred_prob_arry_all = np.transpose(pred_prob_arry_all, (1,0,2,3))

        if flag_use_train == 1:
            np.savez(root_dir+"/train_prob_AR_10.npz", pred_prob_arry_all)
        else:
            np.savez(root_dir+"/test_prob_AR_10.npz", pred_prob_arry_all)

    # ---------------
    if flag_pred2all == 1:

        if flag_use_train == 1:
            pred_path = root_dir+"/y_pred_10"
        else:
            pred_path = root_dir+"/pred_seq_10"
        
        pred_file = [f for f in listdir(pred_path) if isfile(join(pred_path, f))]
        tmp = np.load(pred_path+"/"+pred_file[0])
        pred_arry_all = np.zeros([tmp.shape[0], 121, 240])

        for file in pred_file:
            i = int(file.split("by")[0].split("_")[-1])
            j = int(file.split("by")[1].split("_")[0])
            pred_arry = np.load(pred_path+"/"+file)
            pred_arry_all[:,i,j] = pred_arry

        if flag_use_train == 1:
            np.savez(root_dir+"/train_pred_AR_10.npz", pred_arry_all)
        else:
            np.savez(root_dir+"/test_pred_AR_10.npz", pred_arry_all)

    # ------------------------------

    if flag_pred2oneHot == 1:

        if flag_use_train == 1:
            pred_HMM = np.load(root_dir+"/train_pred_AR_10.npz")["arr_0"]
        else:
            pred_HMM = np.load(root_dir+"/test_pred_AR_10.npz")["arr_0"]

        oneHot_HMM = np.zeros([3, pred_HMM.shape[0], pred_HMM.shape[1], pred_HMM.shape[2]])
        oneHot_HMM[oneHot_HMM == 0] = 'nan'

        for idx in range(len(mask_flt[0])):

            i = mask_flt[0][idx]
            j = mask_flt[1][idx]  

            ds_seq = pred_HMM[:,i,j]
            
            if -1 not in ds_seq:
                ds_seq = np.append(ds_seq, -1)
            if 0 not in ds_seq:
                ds_seq = np.append(ds_seq, 0)
            if 1 not in ds_seq:
                ds_seq = np.append(ds_seq, 1)

            onehot_encoder = OneHotEncoder(sparse=False)
            integer_encoded = np.array(ds_seq).reshape(len(ds_seq), 1)

            onehot_encoded = onehot_encoder.fit_transform(integer_encoded)
            onehot_encoded = onehot_encoded[:pred_HMM.shape[0], :]
            onehot_encoded = np.transpose(onehot_encoded)

            oneHot_HMM[:,:,i,j] = onehot_encoded

        if flag_use_train == 1:
            np.savez(root_dir+"/train_oneHot_HMM.npz", oneHot_HMM)
        else:
            np.savez(root_dir+"/test_oneHot_HMM.npz", oneHot_HMM)

    # ------------------------------

    if flag_get_RPSS == 1:

        if flag_pred2oneHot == 1:
            if flag_use_train == 1:
                oneHot_HMM = np.load(root_dir+"/train_oneHot_HMM.npz")["arr_0"]
            else:
                oneHot_HMM = np.load(root_dir+"/test_oneHot_HMM.npz")["arr_0"]
        if flag_prob2all == 1:
            if flag_use_train == 1:
                oneHot_HMM = np.load(root_dir+"/train_prob_AR_10.npz")["arr_0"]
            else:
                oneHot_HMM = np.load(root_dir+"/test_prob_AR_10.npz")["arr_0"]

        if flag_use_train == 1:
            yr_len = oneHot_HMM.shape[1]//53
            strt = int(time_train_end)-yr_len+1
            time = slice(str(strt), time_train_end)
            ds_cmpr=hind_2000_2019_w34
        else:
            # time=slice(time_valid_start, time_valid_end)
            # ds_cmpr=hind_2000_2019_w34
            time='2020'
            ds_cmpr=fct_2020_w34

        oneHot_HMM_p = func_get_RPSS(oneHot_HMM, ds_cmpr, time)

        preds = oneHot_HMM_p
        xr.set_options(keep_attrs=True)

        if flag_use_train == 1:
            try:
                obs_p = xr.open_dataset(f'{cache_path}/hindcast-like-observations_2000-2019_biweekly_terciled.nc').sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)
            except:
                obs_p = xr.open_zarr(f"{cache_path}/hindcast-like-observations_2000-2019_biweekly_terciled.zarr").sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)
        else:
            obs_p = xr.open_dataset(f'{cache_path}/forecast-like-observations_2020_biweekly_terciled.nc').sel(forecast_time=preds.forecast_time, lead_time=preds.lead_time)

        obs_p = obs_p[v].to_dataset(name=v)

        # ML probabilities
        fct_p = preds

        # climatology
        clim_p = xr.DataArray([1/3, 1/3, 1/3], dims='category', coords={'category':['below normal', 'near normal', 'above normal']}).to_dataset(name=v)

        assert_predictions_2020(obs_p)
        assert_predictions_2020(fct_p)

        ## RPSS
        # rps_ML
        rps_ML = xs.rps(obs_p, fct_p, category_edges=None, dim=[], input_distributions='p').compute()
        # rps_clim
        rps_clim = xs.rps(obs_p, clim_p, category_edges=None, dim=[], input_distributions='p').compute()
        
        expect = obs_p.sum('category')
        expect = expect.where(expect > 0.98).where(expect < 1.02)  # should be True if not all NaN
        rps_ML = rps_ML.where(expect, other=2)  # assign RPS=2 where value was expected but NaN found

        ### rpss
        rpss = 1 - (rps_ML.groupby('forecast_time.year').mean() / rps_clim.groupby('forecast_time.year').mean())
        # dirty fix
        rpss = rpss.clip(-10, 1)
        rrr = rpss.to_array().values.squeeze()

        weights = np.cos(np.deg2rad(np.abs(rpss.latitude)))
        
        loc_region = d_region[name]

        u1, d1, l1, r1 = loc_region[0]
        d1, r1 = d1-1, r1-1
        l1_m, r1_m = l1*1.5, r1*1.5
        u1_m, d1_m = u1*(-1.5)+90, d1*(-1.5)+90
        part = rpss.sel(latitude=slice(u1_m, d1_m), longitude=slice(l1_m, r1_m))# .weighted(weights).mean('latitude').mean('longitude')

        if len(loc_region) > 1:
            u2, d2, l2, r2 = loc_region[1]
            d2, r2 = d2-1, r2-1
            l2_m, r2_m = l2*1.5, r2*1.5
            u2_m, d2_m = u2*(-1.5)+90, d2*(-1.5)+90
            part2 = rpss.sel(latitude=slice(u2_m, d2_m), longitude=slice(l2_m, r2_m))#.weighted(weights).mean('latitude').mean('longitude')
            part = xr.concat([part, part2], dim="longitude")
        
        scores  = part.weighted(weights).mean('latitude').mean('longitude')
        scores = scores.to_array().mean(['lead_time', 'variable'])
   
        df = pd.DataFrame({root_dir.split("/")[-1].split("w560_")[-1]: scores})
        dd = pd.concat([dd, df], axis=1)
    
    dict_path = root_d+"score_dict/"
    if not os.path.exists(dict_path):
        os.makedirs(dict_path)
    dd.to_csv(dict_path+name+"_score_df.csv")


print(">>>>>>>>>> Finished <<<<<<<<<<<")